import os
import sys
import json
import copy
import tempfile
import shutil
import subprocess
import numpy as np
import pandas as pd
from enrichsdk import Compute, S3Mixin
from datetime import datetime
from dateutil import parser as dateparser, relativedelta
import logging

logger = logging.getLogger("app")

from .lib import *

class QueryExecutorBase(Compute):
    """
    Base class for a QueryExecutor transform. This is useful
    to run queries against backends such as backends such as
    mysql

    Features of transform baseclass include:

        * Support query engines (MySQL, Hive, Presto)
        * Support templatized execution
        * Support arbitrary number of queries
        * Supports a generator function to generate per-interval queries

    Configuration looks like::

        ...
        "args": {
            "cleanup": False,
            "force": True,
            "names": "all",
            "start": "2020-08-01",
            "end": "2020-08-03",
        }

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "QueryExecutorBase"
        self.description = "Execute queries against backends"
        self.supported_extra_args = [
            {
                "name": "names",
                "description": "names of the queries to execute",
                "default": "all",
                "required": False
            },
            {
                "name": "force",
                "description": "Force execution",
                "default": "False",
                "required": False
            },
            {
                "name": "start",
                "description": "Start of the time window",
                "default": "",
                "required": True
            },
            {
                "name": "end",
                "description": "End of the time window",
                "default": "",
                "required": True
            }
        ]

        self.testdata = {
            'data_root': os.path.join(os.environ['ENRICH_TEST'],
                                      self.name),
            'statedir': os.path.join(os.environ['ENRICH_TEST'],
                                     self.name, 'state'),
            'conf': {
                'args': {

                }
            },
            'data': {
            }
        }

    @classmethod
    def instantiable(cls):
        return False

    def preload_clean_args(self, args):
        """
        Check validity of the args
        """
        args = super().preload_clean_args(args)

        if (("start" not in args) or
            ("end" not in args)):
            raise Exception("Start or end of timeframe missing")

        try:
            start = dateparser.parse(args['start'])
            args['start'] = start
            end = dateparser.parse(args['end'])
            args['end'] = end
        except:
            logger.exception("Invalid start or end",
                             extra={
                                 'transform': self.name
                             })
            raise Exception("Invalid start/end datetime specified")

        if (('names' not in args) or
            (not isinstance(args['names'], str)) or
            (len(args['names']) == 0)):
            raise Exception("Invalid list of query names specified")

        # Include force
        force = str(args['force']).lower().strip()
        force = force == 'true'
        args['force'] = force

        # Clean the list of names...
        names = args['names'].split(",")
        names = [n.strip() for n in names if len(n.strip()) > 0]
        args['names'] = [n for n in names if len(n) > 0]

        return args


    ########################################
    # Specification Handling...
    ########################################
    def get_spec(self):
        """
        Get query execution specification. Override this

        Returns:

           specs (list): A list of dictionaries. Each dictionary
                         specifies name, credentials, queries to run

        Example::

           return [
               {
                   "name": "roomdb",
                   "cred": "roomdb",
                   "queries": [
                       {
                           "name": "select_star",
                           "output": "%(data_root)s/shared/db/select_star/%(dt)s.tsv",
                           "sql": "%(transform_root)s/SQL/select_star.sql",
                           "params": {
                            "alpha": 22
                           }
                       }
                   ]
               },
               {
                   "enable": False,
                   "name": "hive",
                   "cred": "hiveserver",
                   "queries": [
                       {
                           "name": "employees",
                           "output": "%(data_root)s/shared/db/employee/%(dt)s.tsv",
                           "sql": "%(transform_root)s/SQL/employees.hql",
                       }
                   ]
               }
           ]

        """
        return []

    def validate_spec(self, spec):
        """
        Check whether specification is valid
        """

        if not isinstance(spec, list):
            raise Exception("Query specification should be a list")


        for specitem in spec:

            if not isinstance(specitem, dict):
                logger.error("Query specification items should be dicts",
                             extra={
                                 'transform': self.name,
                                 'data': "Found {}\n{}".format(str(type(specitem)), specitem)
                             })
                raise Exception("Invalid specification")

            if 'executor' in specitem:
                executor = specitem['executor']
                if ((not callable(executor)) and
                    (not ((isinstance(executor, str)) and
                          (len(executor) > 0) and
                          (not hasattr(self, executor))))):
                    raise Exception("Invalid executor specified: {}".format(executor))

            if 'generator' in specitem:
                generator = specitem['generator']
                if ((not callable(generator)) and
                    (not ((isinstance(generator, str)) and
                          (len(generator) > 0) and
                          (not hasattr(self, generator))))):
                    raise Exception("Invalid generator specified: {}".format(generator))

            expected = ['name', 'cred', 'queries']
            missing = [name for name in expected if name not in specitem]
            if len(missing) > 0:
                logger.error("Query specification items should have required elements",
                             extra={
                                 'transform': self.name,
                                 'data': "Missing {} in:\n{}".format(missing, specitem)
                             })
                raise Exception("Invalid specification")

            # => check the cred
            try:
                cred = self.get_credentials(specitem['cred'])
            except:
                logger.exception("Unable to find credentials",
                                 extra={
                                     'transform': self.name
                                 })
                raise Exception("Invalid specification")

            for q in specitem['queries']:
                if ((not isinstance(q, dict)) or (len(q) == 0)):
                    logger.error("Empty or invalid query specification",
                                 extra={
                                     'transform': self.name,
                                     'data': "Query: {}".format(q)
                                 })
                    raise Exception("Invalid specification")

                expected = ['name', 'sql', 'output']
                missing = [name for name in expected if name not in q]
                if len(missing) > 0:
                    logger.error("Query specification items should have required elements",
                                 extra={
                                     'transform': self.name,
                                     'data': "Missing {} in:\n{}".format(missing, specitem)
                                 })
                    raise Exception("Invalid specification")

                if not isinstance(q['sql'], str) or len(q['sql']) == 0:
                    logger.error("Query specification items has invalid sql",
                                 extra={
                                     'transform': self.name,
                                     'data': "SQL: {}".format(q['sql'])
                                 })
                    raise Exception("Invalid specification")

                if 'generator' in q:
                    generator = q['generator']
                    if ((not callable(generator)) and
                        (not ((isinstance(generator, str)) and
                              (len(generator) > 0) and
                              (not hasattr(self, generator))))):
                        raise Exception("Invalid generator specified: {}".format(generator))

            if 'definitions' in specitem:
                definitions = specitem['definitions']
                if ((not instance(definitions, dict)) or
                    (len(definitions) == 0)):
                    logger.error("Query specification items should have valid definition",
                             extra={
                                 'transform': self.name,
                                 'data': "Expected non-empty dict. Found:\n{}".format(specitem)
                             })
                    raise Exception("Invalid specification")

                available_names = [q['name'] for q in specitem['queries']]
                for k, v in definitions.items():
                    if ((not isinstance(v, list)) or
                        (len(v) == 0)):
                        logger.error("Query specification items should have valid definition",
                                     extra={
                                         'transform': self.name,
                                         'data': "Expected valid non-empty value. Found:\n{}".format(definition)
                                     })
                        raise Exception("Invalid specification")
                    missing = [name for name in v if name not in available_names]
                    if len(missing) > 0:
                        logger.error("Query specification items should have valid definition",
                                     extra={
                                         'transform': self.name,
                                         'data': "Missing: {}\n{}".format(missing, definition)
                                     })
                        raise Exception("Invalid specification")


        # Last check whether the requirements can be satisfied
        names = self.args['names']

        available_names = ['all']
        for specitem in spec:
            if 'definitions' in specitem:
                available_names.extend(list(specitem['definitions'].keys()))
            for q in specitem['queries']:
                available_names.append(q['name'])

        missing = [name for name in names if name not in available_names]
        if len(missing) > 0:
            logger.error("Invalid names in args",
                         extra={
                             'transform': self.name,
                             'data': "Missing: {}\nAvailable: {}".format(missing, available_names)
                        })
            raise Exception("Invalid specification")

    #############################################################
    # Output
    #############################################################
    def get_output_handler(self, query, params):
        """
        Find a handler for the output of the query. This function
        should be over-ridden to compute the handler dynamically.

        """
        if 'output' not in query:
            raise Exception("Unable to determine output handler. No 'output' in query")

        if isinstance(query['output'], str):
            return FileOutputHandler(self, query['output'])

            raise Exception("Unable to determine output. No 'output' in query")

    #############################################################
    # Executors..
    #############################################################
    def get_executor(self, specitem, query, credentials):
        """
        Get executor for a specitem and credentials. This executor
        runs the query.

        The executor could be specified within the query,
        spec, or could default to built-in one based on the
        credentials and dbtype within.

        Args:

            spec (dict): Specification of the query
            query (dict): Particular query to execute
            credentials (dict): Credentials for the backend

        Returns:

            a callable executor
        """

        default_executor = None
        if credentials['dbtype'] == 'mysql':
            default_executor = self.mysql_executor
        elif credentials['dbtype'] == 'hive':
            default_executor = self.hive_executor

        # Executor can be per query or for the entire set
        executor = query.get('executor',
                             specitem.get('executor',
                                          default_executor))

        if (executor is not None) and callable(executor):
            return executor

        if (executor is not None) and hasattr(self, executor):
            return getattr(self, executor)

        raise Exception("Cant find executor: {}".format(executor))

    def mysql_executor(self, specitem, credentials, query, params):
        """
        Built in executor for queries against a mysql backend. The output
        is dumped to a temporary file and then an output handler is called
        for post-processing.

        Args:

            spec (dict): Specification of the query
            query (dict): Particular query to execute
            credentials (dict): Credentials for the backend

        """
        try:

            targetdir = None

            # Should this be forced?
            force   = self.args.get('force', False)
            cleanup = self.args.get('cleanup', True)

            # Get the output filename (s3, hdfspath etc.)
            handler = self.get_output_handler(query, params)

            if (not force) and handler.exists(params):
                logger.debug("[Skipping:Exists] {} {}".format(query['name'],
                                                              params['dt']),
                             extra={
                                 'transform': self.name
                             })
                return

            logger.debug("[Computing] {} {}".format(query['name'],
                                                    params['dt']),
                         extra={
                             'transform': self.name
                         })


            # Create a temp directory
            targetdir = tempfile.mkdtemp(prefix="query_executor_")

            # Process the credentials
            config = get_mysql_config(credentials)

            # Create the environment file
            cnfname = os.path.join(targetdir, "env.sh")
            with open(cnfname, 'w') as fd:
                fd.write("[client]\n")
                for var in ['host', 'user', 'password']:
                    fd.write('{}={}\n'.format(var, config[var]))

            # Instantiate the sql
            sqlfile = self.get_file(query['sql'])
            if not os.path.exists(sqlfile):
                raise Exception("Invalid sql file: {}".format(sqlfile))
            sql = open(sqlfile).read()

            # Resolve the sql content
            sql = sql.format(**params)

            sqlname = os.path.join(targetdir, "run.sql")
            with open(sqlname, 'w') as fd:
                fd.write(sql)

            #=> Now write the script
            tmpname = os.path.join(targetdir, "output.tsv")

            cmd = 'mysql --defaults-extra-file={}'.format(cnfname)

            # Generate the script to run
            script = """#!/bin/bash\n\n"""
            script += "date\n"
            script += "{} -B < {} > {}\n".format(cmd, sqlname, tmpname)
            script += "date\n"
            script += "[ -s {0} ] && sed -i 's/\\r//g' {0}\n".format(tmpname)
            scriptname = os.path.join(targetdir, 'run.sh')
            with open(scriptname, 'w') as fd:
                fd.write(script)

            try:
                process = subprocess.Popen(['/bin/bash', scriptname],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
                out, err = process.communicate()
                logger.debug("Executed the script",
                             extra={
                                 'transform': self.name,
                                 "data": "Output:\n----\n" + out.decode('utf-8') + \
                                         "\n\nError:\n-----\n" + err.decode('utf-8')
                             })
            except:
                logger.exception("Error while executing the script",
                                 extra={
                                     'transform': self.name,
                                 })

            # => Now post-process it..
            handler.process(tmpname, params)

        except:
            logger.exception("Failed to execute",
                             extra={
                                 'transform': self.name
                             })

        try:
            if cleanup and (targetdir is not None) and os.path.exists(targetdir):
                shutil.rmtree(targetdir)
            else:
                logger.warning("Targetdir not removed",
                            extra=self.config.get_extra({
                                'transform': self.name,
                                'data': 'Targetdir: {}'.format(targetdir)
                            }))
        except:
            logger.exception("Cleanup failed",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': 'Targetdir: {}'.format(targetdir)
                             }))


    def hive_executor(self, specitem, credentials, query, params):
        """
        Built in executor for queries against a hive backend. The output
        is dumped to a temporary file and then an output handler is called
        for post-processing.

        Args:

            spec (dict): Specification of the query
            query (dict): Particular query to execute
            credentials (dict): Credentials for the backend

        """
        try:

            targetdir = None

            # Should this be forced?
            force   = self.args.get('force', False)
            cleanup = self.args.get('cleanup', True)

            # Get the output filename (s3, hdfspath etc.)
            handler = self.get_output_handler(query, params)

            if (not force) and handler.exists(params):
                logger.debug("[Skipping:Exists] {} {}".format(query['name'],
                                                              params['dt']),
                             extra={
                                 'transform': self.name
                             })
                return

            logger.debug("[Computing] {} {}".format(query['name'],
                                                    params['dt']),
                         extra={
                             'transform': self.name
                         })


            # Create a temp directory
            targetdir = tempfile.mkdtemp(prefix="query_executor_")

            # Process the credentials
            config = get_mysql_config(credentials)

            # Instantiate the sql
            sqlfile = self.get_file(query['sql'])
            if not os.path.exists(sqlfile):
                raise Exception("Invalid sql file: {}".format(sqlfile))
            sql = open(sqlfile).read()

            # Resolve the sql content
            sql = sql.format(**params)

            sqlname = os.path.join(targetdir, "run.sql")
            with open(sqlname, 'w') as fd:
                fd.write(sql)

            #=> Now write the script
            tmpname = os.path.join(targetdir, "output.tsv")

            cmd = "beeline -u jdbc:hive2://%(host)s:%(port)s --silent=true --verbose=False --outputformat=tsv" % config

            if 'user' in config:
                cmd += " -n '{}'".format(user)

            if 'password' in config:
                cmd += " -p '{}'".format(password)

            # Generate the script to run
            script = """#!/bin/bash\n\n"""
            script += "date\n"
            script += "{} -f {} > {}\n".format(cmd, sqlname, tmpname)
            script += "date\n"
            script += "[ -s {0} ] && sed -i 's/\\r//g' {0}\n".format(tmpname)
            scriptname = os.path.join(targetdir, 'run.sh')
            with open(scriptname, 'w') as fd:
                fd.write(script)

            try:
                process = subprocess.Popen(['/bin/bash', scriptname],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
                out, err = process.communicate()
                logger.debug("Executed the script",
                             extra={
                                 'transform': self.name,
                                 "data": "Output:\n----\n" + out.decode('utf-8') + \
                                         "\n\nError:\n-----\n" + err.decode('utf-8')
                             })
            except:
                logger.exception("Error while executing the script",
                                 extra={
                                     'transform': self.name,
                                 })

            # => Now post-process it..
            handler.process(tmpname, params)

        except:
            logger.exception("Failed to execute",
                             extra={
                                 'transform': self.name
                             })

        try:
            if cleanup and (targetdir is not None) and os.path.exists(targetdir):
                shutil.rmtree(targetdir)
            else:
                logger.warning("Targetdir not removed",
                            extra=self.config.get_extra({
                                'transform': self.name,
                                'data': 'Targetdir: {}'.format(targetdir)
                            }))
        except:
            logger.exception("Cleanup failed",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': 'Targetdir: {}'.format(targetdir)
                             }))


    #########################################################
    # Generators of parameters
    #########################################################
    def get_generator(self, specitem, query):
        """
        Parameters generator. This is useful when a templatized
        query has to be run against the backend over many
        days. The output of the generator function is a list of
        dictionaries each of which is a key-value set for one
        time window (say a day)

        Args:

            spec (dict): Specification of the query
            query (dict): Particular query to execute

        Returns:

            a callable generator function
        """
        generator = query.get('generator',
                              specitem.get('generator',
                                           "generator_daily"))
        if (generator is not None) and callable(generator):
            return generator

        if (generator is not None) and hasattr(self, generator):
            return getattr(self, generator)

        raise Exception("Could not find generator: {}".format(generator))

    def generator_daily(self, spec, specitem, query):
        """
        Built-in function to generate a list of dates (one for each day)
        between two dates.

        """

        start = self.args['start']
        end = self.args['end']
        if start > end:
            start, end = end, start

        if isinstance(start, datetime):
            start = start.date()
        if isinstance(end, datetime):
            end = end.date()

        # Pass any extra parameters
        extra = query.get('params', {})
        paramlist = []
        dt = start
        while dt < end:
            params = {
                'dt': dt.isoformat()
            }
            params.update(extra)

            dt += relativedelta.relativedelta(days=1)
            paramlist.append(params)

        return paramlist

    def process_specitem(self, spec, specitem, toexecute):

        logger.debug("Executing {}".format(specitem['name']),
                     extra={
                         'transform': self.name,
                         'data': "To Execute: " + ", ".join(toexecute)
                     })

        # What should we be talking to?
        cred = self.get_credentials(specitem['cred'])

        nature = cred.get('nature', "unknown")
        dbtype = cred.get('dbtype',"unknown")

        # => Sanity check the credentials
        if ((nature not in ['db']) or (dbtype not in ['mysql', 'hive'])):
            logger.error("Invalid credentials",
                         extra={
                             'transform': self.name,
                             'data': "Nature should be 'db' and dbtype should be 'mysql'"
                         })
            raise Exception("Invalid credentials")


        for query in specitem['queries']:
            if query['name'] in toexecute:
                # Get how to executor the query...
                executor = self.get_executor(specitem, query, cred)
                generator = self.get_generator(specitem, query)
                paramlist = generator(spec, specitem, query)
                for params in paramlist:
                    try:
                        executor(specitem=specitem,
                                 credentials=cred,
                                 query=query,
                                 params=params)
                    except:
                        logger.exception("Failed to execute: {} {}".format(query['name'],
                                                                           params['dt']),
                                         extra={
                                             'transform': self.name
                                         })


    def process_spec(self, spec):
        """
        Process query specification
        """

        names = self.args['names']
        for specitem in spec:

            try:

                itemname = specitem['name']

                enable = specitem.get('enable', True)
                if not enable:
                    logger.error("Skipping: {}. Not enabled.".format(itemname),
                                 extra={
                                     'transform': self.name
                                 })
                    continue

                # What should we be executing to begin with..?
                toexecute = []
                for name in names:
                    if name in specitem.get('definitions', {}):
                        toexecute.extend(specitem['definitions'][name])

                for q in specitem['queries']:
                    if ((name == 'all') or (q['name'] == name)):
                        toexecute.append(q['name'])

                # Cleanup
                toexecute = list(set(toexecute))

                if len(toexecute) == 0:
                    logger.error("No parameter list generated: {}".format(itemname),
                                 extra={
                                     'transform': self.name
                                 })
                    continue

                # Now process the list of queries. Params will be
                # generated per specitem.
                self.process_specitem(spec, specitem, toexecute)

                logger.debug("Completed execution: {}".format(itemname),
                             extra={
                                 'transform': self.name
                             })
            except:
                logger.exception("Unable to execute: {}".format(itemname),
                                 extra={
                                     'transform': self.name
                                 })

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug("Starting Query Execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        start = self.args['start']
        end = self.args['end']

        # Get and validate spec
        spec = self.get_spec()
        self.validate_spec(spec)
        try:
            self.process_spec(spec)
        except:
            logger.exception("Failed while processing spec",
                             extra={
                                 'transform': self.name
                             })
            raise

        logger.debug("Completed Query Execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))
        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass




import os
import sys
import json
from enrichsdk import Source
import logging

logger = logging.getLogger("app")

class JSONSource(Source):
    """
    Load a file into a 'dict' frame in the state.

    Params are meant to be passed as parameter to update_frame.

    Example configuration::

         ...

         "args": {
             "source": {
                 'hello': {
                     'frametype': 'dict',
                     'filename': '%(data_root)s/shared/hello.json',
                     'params': {}
                 }
             }
         }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "JSONSource"

        self.testdata = {
            'conf': {
                'args': {
                    'source': {
                        'hello': {
                            'filename': '%(data_root)s/shared/hello.json',
                            'frametype': 'dict',
                            'params': {
                            }
                        }
                    }
                }
            },
            'data': {
            }
        }


    def preload_clean_args(self, args):
        """
        Check if the args are consistent with the
        specification.
        """

        # Backward compatabilty
        if 'source' not in args:
            args = {
                'source': args
            }

        args = super().preload_clean_args(args)

        # Sanity check...
        assert isinstance(args, dict)
        assert 'source' in args
        assert isinstance(args['source'], dict)

        source = args['source']
        for name, detail  in source.items():

            if (("frametype" not in detail) or
                (detail['frametype'] != 'dict')):
                logger.error("Invalid configuration. Only JSON source supported by this source transform",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            if (('filename' not in detail) or
                (not isinstance(detail['filename'], str)) or
                ('params' not in detail) or
                (not isinstance(detail['params'], dict))):
                logger.error("Invalid args. Filename (string) and params (dict) are required",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            filename = detail['filename']
            if not filename.lower().endswith('.json'):
                logger.error("Input file must a .json file",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            #=> Materialize the path...
            detail['filename'] = self.config.get_file(detail['filename'])

        return args

    def validate_args(self, what, state):
        """
        Double check the arguments
        """
        assert isinstance(self.args, dict)
        assert 'source' in self.args
        assert isinstance(self.args['source'], dict)

        source = self.args['source']
        for name, detail  in source.items():
            assert (('frametype' in detail) and (detail['frametype'] == 'dict'))
            assert 'filename' in detail
            assert 'params' in detail

    def process(self, state):
        """
        Load the json files into 'dict' frames and store them in the state.
        """
        logger.debug("{} - process".format(self.name),
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        source = self.args['source']
        for name, detail  in source.items():

            filename = detail['filename']
            data = json.load(open(filename))

            updated_detail = {
                'df': data,
                'transform': self.name,
                'frametype': 'dict',
                'params': [
                    {
                        'type': 'compute',
                    },
                    {
                        'type': 'lineage',
                        'transform': self.name,
                        'dependencies': [
                            {
                                'type': 'file',
                                'nature': 'input',
                                'objects': [
                                    filename
                                ]
                            }
                        ]
                    }
                ],
                'history': [
                    # Add a log entry describing the change
                    {
                        'transform': self.name,
                        'log': 'Loaded json file'
                    }
                ]
            }


            # Update the state.
            state.update_frame(name, updated_detail, create=True)

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """

        source = self.args['source']
        for name, detail  in source.items():
            if not state.reached_stage(name, self.name):
                raise Exception("Could not find new frame created for {}".format(name))

                detail = state.get_frame(name)
                df = detail['df']

                # Check if it is a valid dictionary...
                assert isinstance(df, dict)


provider = JSONSource



"""
etl_job.py
~~~~~~~~~~

This Python module contains an example Apache Spark ETL job definition
that implements best practices for production ETL jobs. It can be
submitted to a Spark cluster (or locally) using the 'spark-submit'
command found in the '/bin' directory of all Spark distributions
(necessary for running any Spark job, locally or otherwise). For
example, this example script can be executed as follows,

    $SPARK_HOME/bin/spark-submit \
    --master spark://localhost:7077 \
    --py-files packages.zip \
    --files configs/etl_config.json \
    jobs/etl_job.py

where packages.zip contains Python modules required by ETL job (in
this example it contains a class to provide access to Spark's logger),
which need to be made available to each executor process on every node
in the cluster; etl_config.json is a text file sent to the cluster,
containing a JSON object with all of the configuration parameters
required by the ETL job; and, etl_job.py contains the Spark application
to be executed by a driver process on the Spark master node.

For more details on submitting Spark applications, please see here:
http://spark.apache.org/docs/latest/submitting-applications.html

Our chosen approach for structuring jobs is to separate the individual
'units' of ETL - the Extract, Transform and Load parts - into dedicated
functions, such that the key Transform steps can be covered by tests
and jobs or called from within another environment (e.g. a Jupyter or
Zeppelin notebook).
"""

import os
import sys 
import json
import argparse
import traceback

from dependencies import enrich

def add_data(config):

    sqlsc = config['sqlsc'] 
    root = config['root'] 
    data = {} 
    for name, path in config['data'].items():
        filename = os.path.join(root, path) 
        df = sqlsc.read.csv(filename,
                            header=True, 
                            mode="DROPMALFORMED",
                            sep=",")

        
        data[name] = df.rdd

    config['rdds'] = data 

def run(config):

    from dependencies import supernova 

    add_data(config)
    
    params = config['params']
    
    # Initialize the tagger...
    tagging_extra = supernova.initialize(params)
            
    rdd = config['rdds']['members'] 
    def reducefunc(g):
        key = g[0]
        rows = g[1] 
        rows = list(rows)

        # All elements are text. So fix the types..
        
        
        # Now run the tagger..
        records = supernova.process(rows, tagging_extra) 
        
        print("LOOKING AT KEY", key) 
        print("LOOKING AT ROWS", rows)
        result = (key, len(rows))
        print("RESULT", result) 
        return {
            'key': key,
            'len': len(rows) 
        }
    
    def grouping(row):        
        return tuple([row[x] for x in ['MEMBERSHIP_ID',
                                       'bdate',
                                       'storenmbr']])
    grouped = rdd.groupBy(grouping) 
    
    collected = grouped.map(reducefunc)
    print("GROUPED COLLECT", collected.take(2))

    collected = collected.collect()
    
    print("COLLECTED", collected) 


# entry point for PySpark ETL application
if __name__ == '__main__':

    # => Load configuration and pass any additional parameters...
    parser = argparse.ArgumentParser()    
    parser.add_argument('--config',
                        required=True,
                        help='Configuration file') 
    args = parser.parse_args() 
    config = json.load(open(args.config))
    config['root'] = os.path.dirname(args.config)

    #Start the spark session 
    enrich.start_spark_session(config)

    # Run the code...
    try:
        run(config)
    except:
        traceback.print_exc()

    # Shut down 
    enrich.stop_spark_session(config) 



import os
import logging 
from jinja2 import Environment, FileSystemLoader, meta

logger = logging.getLogger('app')

# Template rendering functions
def template_render(widgetdir, widgetname, context):
    """
    Render a specified template using the context 
    
    Args: 
      widgetname (str): Name of the template 
      context (dict): Key-value pairs 

    Returns:
      Rendered html template that can be embedded 

    """
    filename = widgetname + ".html"         
    env = Environment(loader=FileSystemLoader(widgetdir))
    try: 
        template = env.get_template(filename) 
        return template.render(context)
    except:
        logger.exception("Could not find template") 
        return "" 

def template_get_variables(widgetdir, widgetname):     
    """
    Get the variables from a template
    
    Args: 
       widgetname (str): Name of the template 
    
    Returns:
       List of variables
    """
    filename = widgetname + ".html"         
    env = Environment(loader=FileSystemLoader(widgetdir))
    try: 
        template_source = env.loader.get_source(env, filename)[0]
        parsed_content = env.parse(template_source)
        return meta.find_undeclared_variables(parsed_content) 
    except:
        return []



"""
Enrich Platform module packaging service.

This service allows you to boostrap (create a new module) and unit
test it before uploading to the server.

"""

import os
import sys
import json
import stat
import warnings
import imp
import copy
import glob2
import shutil
import importlib
import importlib.util
import traceback
import inspect
import logging
from datetime import datetime, date, timezone, timedelta
from time import gmtime, strftime
from collections import OrderedDict, defaultdict

import pytz
from tzlocal import get_localzone
from dateutil import parser as dateparser

from enrichsdk.tasks import Task
from enrichsdk.tasks.sdk import *
from prompt_toolkit import prompt

try:
  from prompt_toolkit.contrib.completers import WordCompleter
except:
  from prompt_toolkit.completion import WordCompleter

from .validators import *
from .lib import *
from .log import setup_logging
from .mock import *
from ..lib import *
from ..lib.customer import load_customer_assets
from ..app.flask.render import create_app
from enrichsdk.tasks import Task

__all__ = ['bootstrap', 'test_transform', 'test_conf',
           'test_task', 'checkenv', 'print_log',
           'testdata_action',
           'get_sample_siteconf', 'get_sample_versionmap',
           'populate']

class DataFilter(logging.Filter):
    def filter(self, record):

        data = getattr(record, 'data', '')
        if isinstance(data, str) and len(data) > 0:
            data = "\n"+data
        record.data = data

        transform = getattr(record, 'transform', '')
        record.transform = transform

        return True

# Insert the filter for app messages
logger = logging.getLogger('app')
df = DataFilter()
logger.addFilter(df)
for name in [
        'kazoo.client',
        'pykafka.simpleconsumer',
        'pykafka.handlers',
        'pykafka.balancedconsumer',
        'pykafka.connection',
        'pykafka.membershipprotocol',
        'pykafka.cluster',
        'pykafka.topic'
]:
    logging.getLogger(name).disabled = True


#################################################
# Helper
#################################################
def welcome():
    print("==============================")
    print("  Welcome to Scribble Enrich ")
    print("=============================")
    print("\n")

def read_conf():

    # Created a global configuration
    confpath = "enrich.json"
    if os.path.exists(confpath):
        try:
            conf = json.loads(open(confpath).read())
            assert "org" in conf
            assert "repository" in conf
            return conf
        except:
            traceback.print_exc()
            pass

    print("First we have to create a configuration file (enrich.json). After we update you")
    print("will find the content in {}".format(confpath))

    msg = """\nPlease enter your customer organization details.

Note that the organization name has to be precise. Please check with
Scribble if you are not sure\n"""
    print(msg)

    validators = OrderedDict()
    defaults = {
        'name': os.path.basename(os.path.abspath("."))
    }

    validators['name'] = NameValidator()
    validators["description"] =  OrgDescValidator()
    validators["logourl"] =  LogoValidator()

    orgcontext = {}
    for v in validators:
        orgcontext[v] = prompt('> ' + v +": ",
                               default=defaults.get(v,""),
                               validator=validators.get(v,None))

    msg = """\nThanks! \n\nNow, please enter your and details of this code repository\n"""
    print(msg)

    validators = OrderedDict()
    validators["name"] =  NameValidator()
    validators["email"] =  EmailValidator()
    validators["giturl"] =  URLValidator()

    devcontext = {}
    for v in validators:
        devcontext[v] = prompt('> ' + v +": ",
                            validator=validators.get(v,None))

    devcontext['author'] = devcontext.pop('name')

    context = orgcontext
    context.update(devcontext)

    # Get the template
    template = get_template_path('enrich.json.template')

    # Write a local copy
    content = write_rendered_file(template, ".",
                                  'enrich.json', context)

    # Write a global enrich.json as well
    content = write_rendered_file(template,
                                  os.environ['ENRICH_ETC'],
                                  'enrich.json', context)

    print("Wrote global configuration file {}".format(confpath))

    return json.loads(content)

############################################################
# Prepare workspace
############################################################
def prepare_enrich_root(target):
    for d in ['data', 'etc', 'customers']:
        try:
            os.makedirs(os.path.join(target, d))
        except:
            pass

############################################################
# Prepare repo
############################################################
def prepare_repo(target):

    # First create init file
    template = get_template_path('repo.init.template')
    write_rendered_file(template, target, '__init__.py', {})
    print(greenbg("Wrote __init__.py in {}".format(target)))

    # Load if one already exists..
    try:
        defaults = load_enrichjson_defaults(target,usecase=False)
    except:
        traceback.print_exc()
        defaults = {}

    prepare_enrichjson(target,
                       defaults=defaults,
                       template="root.enrich.json.template")
    print(greenbg("Wrote enrich.json in {}".format(target)))

    defaults = load_enrichjson_defaults(target,usecase=False)

    msg = """\
[%(org_name)s] Code Repository
--------

Enrich modules for various usecases""" % defaults
    readme = os.path.join(target, 'README.rst')
    if not os.path.exists(readme):
        with open(readme, 'w') as fd:
            fd.write(msg)
    print(greenbg("Added a README.rst in {}".format(target)))

    requirements = os.path.join(target, 'requirements.txt')
    if not os.path.exists(requirements):
        with open(requirements, 'w') as fd:
            fd.write("#")
    print(greenbg("Added a requirements.txt in {}".format(target)))


############################################################
# Prepare workspace
############################################################
def load_enrichjson_defaults(target, usecase=True):

    defaults = {}

    if usecase:
        enrichjsons = glob2.glob(target + "/../*/enrich.json") + \
                      glob2.glob(target + "/../*enrich.json")
    else:
        enrichjsons = glob2.glob(target + "/*enrich.json")

        # Initialize
        defaults['org_name'] = os.path.basename(os.path.abspath(target))
        defaults['usecase_name'] = "Unknown"


    for enrichjson in enrichjsons:
      try:
        j = json.load(open(enrichjson))
        defaults['org_name'] = j['org']['customer']
        defaults['org_logourl'] = j['org']['logo']
        defaults['org_description'] = j['org']['description']
        defaults['author_name'] = j['repository']['author']
        defaults['author_email'] = j['repository']['email']
        defaults['repo_giturl'] = j['repository']['giturl']
      except:
        pass

    return defaults

def prepare_workspace(target, force=False, auto=False):
    if "/" in target and not force:
        raise Exception("Only sub-directory name should be mentioned without /")

    usecasename = os.path.basename(target)
    customer = os.path.basename(os.path.dirname(target))

    parentenrich = os.path.join(customer, "enrich.json")
    if not os.path.exists(parentenrich):
        raise Exception("enrich.json missing in root directory")

    for d in [
            'pipelines',
            'pipelines/conf',
            'pipelines/lib',
            'pipelines/jobs',
            'dashboard',
            'bin',
            'docs',
            'commands',
            'transforms',
            'assets',
            'tasks',
            'tasks/conf',
            'tasks/lib',
            'tasks/jobs',
            'workflows',
            'workflows/prefect',
            'workflows/spark',
    ]:
        path = os.path.join(target, d)
        try:
            os.makedirs(path)
        except:
            pass
        print("Created {}".format(path))

    try:
        data_root = os.path.join(os.environ['ENRICH_TEST'], usecasename)
        if not os.path.exists(data_root):
          os.makedirs(data_root)
          print("Created data_root in {}".format(data_root))

    except:
        traceback.print_exc()


    msg = """\
[{0}] {1}
--------

Enrich modules for {1}""".format(customer, usecasename)
    readme = os.path.join(target, 'README.rst')
    if not os.path.exists(readme):
        with open(readme, 'w') as fd:
            fd.write(msg)
    print("Added a readme in {}".format(readme))

    defaults = load_enrichjson_defaults(target,usecase=True)

    prepare_enrichjson(target, defaults, auto)


def prepare_enrichjson(target, defaults, auto=False,template='enrich.json.template'):

    print("Preparing enrich.json metadata file")

    enrichjson = "enrich.json"
    filename = os.path.join(target, enrichjson)

    # Get the template
    template = get_template_path(template)

    variables = get_variables(template)

    validators = {}
    validators['usecase_name'] = NameValidator()

    context = {}
    if not auto:
      for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=defaults.get(v,""),
                            validator=validators.get(v,None))
    else:
      for v in variables:
        context[v] = defaults.get(v, 'unspecified')

    # Write a local copy
    write_rendered_file(template, target,
                        'enrich.json',
                        context)



############################################################
# Convert
############################################################
def prepare_setup(component, target):

    if os.path.exists(target):
        raise Exception("Destination exists. Please remove it.")

    conf = read_conf()

    basename = os.path.basename(target)
    repo = conf.get('repository',{})
    org = conf.get('org',{})
    params = {
        'component': component,
        'basename': basename,
        'name': basename,
        'version': '0.1',
        'description': basename + " package",
        'author': repo.get('author',''),
        'email': repo.get('email',''),
        'url': repo.get('giturl',''),
        'customer': org.get('name','')
    }


    # => Write the manifest file
    # Write the manifest file...
    write_rendered_file('MANIFEST.in.template',
                        target, 'MANIFEST.in', params)
    write_rendered_file('setup.cfg.template',
                        target, 'setup.cfg', params)
    write_rendered_file('LICENSE.template',
                        target, 'LICENSE', params)
    write_rendered_file('setup.py.template',
                        target, 'setup.py', params)

    testdir = os.path.join(target, 'tests')
    try:
        os.makedirs(testdir)
    except:
        pass

    write_rendered_file('test_module.py.template',
                        testdir, 'test_module.py',
                        params)

    # Copy the fixtures
    src = os.path.join(default_templatedir, 'fixtures')
    dst = os.path.join(testdir, 'fixtures')
    shutil.copytree(src, dst)

    with open(os.path.join(testdir, '__init__.py'), 'w') as fd:
        fd.write("")

    print(greenbg("Created a package in {}".format(target)))
    print(redbg("Note: Please look at fixtures in tests/fixtures and update the paths for data_root and statedir"))


def prepare_asset(what, target, templatepath):

    print("Bootstrapping a {}".format(what))
    if what not in ['asset']:
        raise Exception("Unsupported component")

    if templatepath is not None:
        raise Exception("Asset template not supported yet")

    if os.path.exists(target):
        raise Exception("Target directory exists. Please remove it")

    conf = read_conf()

    # Create setup.py etc.
    prepare_setup(what, target)

    content = '''\
#!/usr/bin/env python
# coding: utf-8
"""
    A description which can be long and explain the complete
    functionality of this module even with indented code examples.
    Class/Function however should not be documented here. They can
    go into the individual functions.

"""
    '''

    # Initialize the library...
    basename = os.path.basename(target)
    initfile = os.path.join(target,'src', basename, '__init__.py')
    try:
        os.makedirs(os.path.dirname(initfile))
    except:
        pass
    with open(initfile, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped asset library {} in {}".format(basename, target))


def prepare_component(what, target, templatepath):

    print("Bootstrapping a {}".format(what))
    if what not in ['transform-package']:
        raise Exception("Unsupported component")

    if os.path.exists(target):
        raise Exception("Target directory exists. Please remove it")

    conf = read_conf()

    # Create setup.py etc.
    prepare_setup(what, target)

    # => Load the manifest
    print("Please provide component details:")
    defaults = {
        'version': "1.0",
        'minorversion': "1.0",
        "author_name": conf.get('repository',{}).get('author','John Smith'),
        "author_email": conf.get('repository',{}).get('email','john.smith@alphainc.com'),
        "repo_giturl": conf.get('repository',{}).get('giturl','git@github.com:alpha/enrich-alpha.git'),
        "org_logourl": conf.get('org',{}).get('logo',"https://alphainc.com/logo/small.png"),
        "org_name": conf.get('org',{}).get('name',"Alpha Inc"),
        "org_description": conf.get('description',{}).get('name',"Alpha retail"),
    }

    validators = {
        'name': NameValidator(),
        'transform_name': NameValidator(),
        'description': GenericDescValidator(),
        'transform_description': GenericDescValidator()
    }

    manifestfile = get_template_path('manifest.json.template')
    context = {}
    variables = get_variables(manifestfile)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=defaults.get(v,""),
                            validator=validators.get(v,None))

    dst = os.path.join(target, os.path.basename(target))
    write_rendered_file(manifestfile, dst, 'manifest.json', context)
    write_rendered_file('README.md.template', dst, 'README.md', context)

    # Get the right template
    if templatepath is None:
        templates = list_templates(suffix=".node.template")
        templatepath = templates['transform']
        #templatepath = 'transform.node.template'
        #template_completer = WordCompleter(list(templates.keys()),
        #                                   ignore_case=True)
        #name = prompt('> Type of Transform: ',
        #              default=list(templates.keys())[0],
        #              completer=template_completer)
        #templatepath = templates[name]

    variables = get_variables(templatepath)
    for v in variables:
      if v not in context:
        context[v] = prompt('> ' + v +": ",
                          default=defaults.get(v,""),
                            validator=validators.get(v,None))
    write_rendered_file(templatepath, dst, '__init__.py', context)

    # Bye bye
    print("Bootstrapped {} {} in {}".format(what, context['transform_name'], target))

def prepare_config(path, templatepath):

    welcome()

    print("Bootstrapping a new configuration\n")

    conf = read_conf()

    validators = {
        'name': NameValidator(),
        'description': GenericDescValidator()
    }

    print("\nPlease provide configuration details\n")
    if templatepath is None:
        config_template = get_template_path('config.json.template')
    else:
        config_template = templatepath

    context = {}
    variables = get_variables(config_template)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            validator=validators.get(v,None))
    config = render(config_template, context)

    with open(path, 'w') as fd:
        fd.write(config)

    # Bye bye
    print("Bootstrapped configuration {} in {}".format(context['name'], path))

def prepare_tasklib(path, templatepath):

    welcome()

    print("Bootstrapping a new task library\n")

    conf = read_conf()

    validators = {
        'name': NameValidator(),
        'description': GenericDescValidator()
    }

    print("\nPlease provide configuration details\n")
    if templatepath is None:
        lib_template = get_template_path('tasklib.template')
    else:
        lib_template = templatepath

    context = {}
    variables = get_variables(lib_template)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            validator=validators.get(v,None))
    content = render(lib_template, context)

    try:
        os.makedirs(path)
    except:
        pass

    initfile = os.path.join(path, "__init__.py")
    with open(initfile, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped configuration {} in {}".format(context['name'], path))


def prepare_taskconf(path, templatepath):

    welcome()

    print("Bootstrapping a new task configuration file\n")

    conf = read_conf()

    if not path.lower().endswith(".json"):
        print("Task configuration file should be a json")
        return

    validators = {
        'taskdag_name': NameValidator(),
        'taskdag_name': NameValidator(),
        'runid_prefix': NameValidator(),
        'task_name': NameValidator(),
        'description': GenericDescValidator()
    }

    print("\nPlease provide configuration details\n")
    if templatepath is None:
        lib_template = get_template_path('taskconf.template')
    else:
        lib_template = templatepath

    defaults = {
        'customer': conf['org']['customer'],
        'usecase': conf['org']['name'],
    }


    context = {}
    variables = get_variables(lib_template)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=defaults.get(v,""),
                            validator=validators.get(v,None))
    content = render(lib_template, context)

    try:
        os.makedirs(os.path.dirname(path))
    except:
        pass

    with open(path, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped configuration {} in {}".format(context['taskdag_name'], path))

def prepare_pipelineconf(path, templatepath):

    welcome()

    print("Bootstrapping a new pipeline configuration file\n")

    conf = read_conf()

    if not (path.lower().endswith(".json") or
            path.lower().endswith(".py")):
        print("Pipeline configuration file should be a json")
        return

    validators = {
        'pipeline_name': NameValidator(),
        'pipeline_description': GenericDescValidator(),
        'usecase': NameValidator(),
        'runid_prefix': NameValidator(),
        'notification_email': EmailValidator(),
        'transform': NameValidator(),
        'realtime_or_batch': RealtimeBatchValidator(),
    }

    print("\nPlease provide configuration details\n")

    if templatepath is None:
        if path.lower().endswith('.py'):
            lib_template = get_template_path('pipelinepyconf.template')
        else:
            lib_template = get_template_path('pipelineconf.template')
    else:
        lib_template = templatepath

    defaults = {
        'realtime_or_batch': 'batch',
        'customer': conf.get('org',{}).get('customer',''),
        'usecase': conf.get('org',{}).get('name','')
    }

    context = {}
    variables = get_variables(lib_template)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=defaults.get(v,""),
                            validator=validators.get(v,None))

    content = render(lib_template, context)

    try:
        os.makedirs(os.path.dirname(path))
    except:
        pass

    with open(path, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped pipeline specification in {}".format(path))

def prepare_script_default(path, templatepath):

    if os.path.exists(path):
        raise Exception("Output path already exists. Please remove it")

    conf = read_conf()

    validators = {
        'confname': NameValidator(),
        'runname': NameValidator(),
        'email': EmailValidator(),
        'transform': NameValidator(),
        'description': GenericDescValidator()
    }

    context = {
        'today': date.today().isoformat(),
        'author_email': conf['repository']['email'],
        'email': conf['repository']['email'],
        'author': conf['repository']['author'],
        'customer': conf['org']['customer'],
        'division': conf['org']['name'],
    }
    variables = get_variables(templatepath)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=context.get(v, ""),
                            validator=validators.get(v,None))

    content = render(templatepath, context)

    try:
        os.makedirs(os.path.dirname(path))
    except:
        pass

    with open(path, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped script in {}".format(path))

def prepare_script_minimal(path, templatepath):

    if os.path.exists(path):
        raise Exception("Output path already exists. Please remove it")

    validators = {
        'confname': NameValidator(),
        'runname': NameValidator(),
        'email': EmailValidator(),
        'transform': NameValidator(),
        'description': GenericDescValidator()
    }

    context = {
        'today': date.today().isoformat(),
    }

    variables = get_variables(templatepath)
    for v in variables:
        context[v] = prompt('> ' + v +": ",
                            default=context.get(v, ""),
                            validator=validators.get(v,None))

    content = render(templatepath, context)

    try:
        os.makedirs(os.path.dirname(path))
    except:
        pass

    with open(path, 'w') as fd:
        fd.write(content)

    # Bye bye
    print("Bootstrapped script in {}".format(path))

def prepare_rscript(path, templatepath):

    welcome()

    print("Bootstrapping an R Script\n")

    if not path.lower().endswith(".r"):
        print("R script should end with .r/R")
        return

    if templatepath is None:
        lib_template = get_template_path('rscript.template')
    else:
        lib_template = templatepath

    prepare_script_default(path, lib_template)

def prepare_pyscript(path, templatepath):

    welcome()

    print("Bootstrapping an Python Script\n")

    if not path.lower().endswith(".py"):
        print("py script should end with .py")
        return

    if templatepath is None:
        lib_template = get_template_path('pyscript.template')
    else:
        lib_template = templatepath

    prepare_script_default(path, lib_template)

def prepare_transform_script(path, templatepath,what='simpletransform'):

    welcome()

    print("Bootstrapping a simple transform\n")

    path = path.strip()
    if not path.endswith(".py"):
        path = path + "/__init__.py"

    if templatepath is None:
        lib_template = get_template_path(what+'.node.template')
    else:
        lib_template = templatepath

    prepare_script_minimal(path, lib_template)


def prepare_sparkjob(path, templatepath):
    """
    Bootstrap a new sparkjob

    Parameters
    ----------
    component: Nature of the module that should be bootstrapped
    path: Path where the module should be created.

    Yields
    ------
    directory hierarchy

    """
    if templatepath is not None:
        raise Exception("sparkjob template not supported yet")

    src = os.path.join(default_templatedir, 'spark')
    dst = path
    if os.path.exists(dst):
        raise Exception("Destination exists. Please remove it")

    shutil.copytree(src, dst)

    # Bye bye
    print("Bootstrapped sparkjob")

def prepare_airflowjob(path, templatepath):
    """
    Bootstrap a new airflowjob

    Parameters
    ----------
    component: Nature of the module that should be bootstrapped
    path: Path where the module should be created.

    Yields
    ------
    directory hierarchy

    """

    if templatepath is not None:
        raise Exception("airflowjob template not supported yet")

    src = os.path.join(default_templatedir,
                       'airflow',
                       'contrib-pipeline-v1.py')
    dst = path
    if os.path.exists(dst):
        raise Exception("Destination exists. Please remove it")

    shutil.copy(src, dst)

    # Bye bye
    print("Bootstrapped airflowjob")

def prepare_prefectjob(path, templatepath):
    """
    Bootstrap a new prefect

    Parameters
    ----------
    component: Nature of the module that should be bootstrapped
    path: Path where the module should be created.

    Yields
    ------
    directory hierarchy

    """

    if templatepath is not None:
        raise Exception("prefectjob template not supported yet")

    src = os.path.join(default_templatedir,
                       'prefect',
                       'default.py')
    dst = path
    if os.path.exists(dst):
        raise Exception("Destination exists. Please remove it")

    shutil.copy(src, dst)

    # Update the execute permissions
    st = os.stat(dst)
    os.chmod(dst,  st.st_mode | stat.S_IXGRP | stat.S_IXOTH | stat.S_IXUSR)

    # Bye bye
    print("Bootstrapped prefectjob")

def bootstrap(component, path, templatepath):
    """
    Bootstrap a new module (python package)

    Currently transform and model are supported modules

    Parameters
    ----------
    component: Nature of the module that should be bootstrapped
    path: Path where the module should be created.

    Yields
    ------
    A python package

    """

    if False:
        if 'ENRICH_ROOT' not in os.environ:
            print("ENRICH_ROOT variable not defined. Looks like you haven't source enrich-env.sh")
            raise Exception("Environment missing: Please define ENRICH_ROOT")

    path = path.rstrip('/') #For the case when user has added a trailing '/'

    if component == 'transform-package':
        prepare_component('transform-package', path, templatepath)
    elif component in ['transform-simple']:
        prepare_transform_script(path, templatepath,what='simpletransform')
    elif component in ['transform-helloworld']:
        prepare_transform_script(path, templatepath, what='helloworld')
    #elif component == 'model':
    #    prepare_component('model', path, templatepath)
    #elif component == 'skin':
    #    prepare_component('skin', path, templatepath)
    elif component == 'config':
        prepare_config(path, templatepath)
    elif component == 'repo':
        prepare_repo(path)
    elif component == 'asset':
        prepare_asset('asset', path, templatepath)
    elif component in ['tasklib', 'task-lib']:
        prepare_tasklib(path, templatepath)
    elif component in ['taskconf', 'task-conf']:
        prepare_taskconf(path, templatepath)
    elif component in ['pipelineconf', 'pipeline-conf']:
        prepare_pipelineconf(path, templatepath)
    elif component in ['workspace', 'usecase']:
        prepare_workspace(path)
    elif component in ['sparkjob', 'spark', 'spark-job']:
        prepare_sparkjob(path, templatepath)
    elif component in ['airflowjob', 'airflow']:
        prepare_airflowjob(path, templatepath)
    elif component in ['prefect', 'prefectjob']:
        prepare_prefectjob(path, templatepath)
    elif component in ['rscript']:
        prepare_rscript(path, templatepath)
    elif component in ['pyscript']:
        prepare_pyscript(path, templatepath)

def get_sample_siteconf():
    siteconf = """\
{
    "customer": "Acme Inc",
    "dashboard": {
        "title": "Acme Rich Data Platform"
    },
    "credentials": {
        "data-bucket": {
            "nature": "s3",
            "bucket": "acme-datalake",
            "readonly": false,
            "access_key": "AKIAJURXL...Q",
            "secret_key": "tutww...A"
        }
    }
}\
    """
    return siteconf

def get_sample_versionmap():
    versionmap = """\
[
    {
    "description": "Enrich SDK for pipeline development",
    "date": "2018-03-20 16:14:19 +0530",
    "repo": "enrich-sdk",
    "release": "v1.5.5",
    "label": "sdk",
    "commit": "6b406f0b083b962c164ac54431a44811c45b9ff0"
    }
]\
    """
    return versionmap

def populate(context={}):
    """
    Creates the enrich working space
    """
    variables = [
        "ENRICH_ROOT",
        "ENRICH_DATA",
        "ENRICH_TEST",
        "ENRICH_ETC",
        "ENRICH_SHARED",
        "ENRICH_VAR",
        "ENRICH_LIB",
        "ENRICH_OPT",
        "ENRICH_LOGS",
        "ENRICH_CUSTOMERS",
        "ENRICH_RELEASES",
        "ENRICH_TEST",
    ]
    for var in variables:
        if var in os.environ:
            path = os.environ[var]
            try:
                if os.path.exists(path):
                    print(orangebg(var + " exists"))
                else:
                    os.makedirs(path)
                    print(greenbg(var + " created"))
            except:
                pass

    siteconf = None
    if 'siteconf' in os.environ:
        siteconf = os.environ['siteconf']
    elif 'siteconf' in context:
        siteconf = context['siteconf']
    elif 'ENRICH_ETC' in os.environ:
        siteconf = os.path.join(os.environ['ENRICH_ETC'], 'siteconf.json')
    else:
        print(redbg("Path for siteconf not found"))

    if siteconf is not None:
        if os.path.exists(siteconf):
            print(orangebg("siteconf already present"))
        else:
            try:
                with open(siteconf, 'w') as fd:
                    fd.write(get_sample_siteconf())
                print(greenbg("siteconf initialized"))
            except:
                print(redbg("siteconf could not be created"))

    #=> Versionmap
    versionmap = None
    if 'versionmap' in os.environ:
        versionmap = os.environ['versionmap']
    elif 'versionmap' in context:
        versionmap = context['versionmap']
    elif 'ENRICH_ETC' in os.environ:
        versionmap = os.path.join(os.environ['ENRICH_ETC'], 'versionmap.json')
    else:
        print(redbg("Path for versionmap not found"))

    if versionmap is not None:
        if os.path.exists(versionmap):
            print(orangebg("versionmap already present"))
        else:
            try:
                with open(versionmap, 'w') as fd:
                    fd.write(get_sample_versionmap())
                print(greenbg("versionmap initialized"))
            except:
                print(redbg("versionmap could not be created"))

    if 'ENRICH_CUSTOMERS' in os.environ and 'ENRICH_DATA' in os.environ:
        enrich_customers = os.environ['ENRICH_CUSTOMERS']
        enrich_data = os.environ['ENRICH_DATA']
        if os.path.exists(enrich_customers):
            customers = os.listdir(enrich_customers)
            if len(customers) == 0:
                print(redbg("No applications linked in ENRICH_CUSTOMERS {}".format(enrich_customers)))
            else:
                for c in customers:
                    for usecase in os.listdir(os.path.join(enrich_customers, c)):
                        usecase_root = os.path.join(enrich_customers, c, usecase)
                        enrichjson = os.path.join(usecase_root, 'enrich.json')
                        if not os.path.exists(enrichjson):
                            continue
                        data_root = os.path.join(enrich_data, c, usecase)
                        if not os.path.exists(data_root):
                            os.makedirs(data_root)
                            print(greenbg("data_root created: {}".format(data_root)))
                        else:
                            print(orangebg("data_root exists: {}".format(data_root)))

def checkenv(validate_only=False,context={}):
    """
    Check whether environment is correctly configured

    Parameters
    ----------
    None

    Yields
    ------
    Errors if any
    """

    variables = [
        "ENRICH_ROOT",
        "ENRICH_DATA",
        "ENRICH_TEST",
        "ENRICH_ETC",
        "ENRICH_SHARED",
        "ENRICH_VAR",
        "ENRICH_LIB",
        "ENRICH_OPT",
        "ENRICH_LOGS",
        "ENRICH_CUSTOMERS",
        "ENRICH_RELEASES",
        "ENRICH_CUSTOMERS",
        "ENRICH_TEST",

        # Tasks
        #"ENRICH_TASKDB_URL",
        #"ENRICH_TASKDB_HOSTNAME",
        #"ENRICH_TASKDB_PORT",
        #"ENRICH_TASKDB_LOGFILE",

        # Email
        #"EMAIL_HOST",
        #"EMAIL_HOST_USER",
        #"EMAIL_HOST_PASSWORD",
        #"EMAIL_PORT",
        #"EMAIL_USE_TLS"

    ]

    if not validate_only:
        print("Checking Environment")
        print("===============\n")

    for var in variables:
        if var in os.environ:
            if not validate_only:
                print(greenbg("{} defined".format(var)))
        else:
            if validate_only:
                return False
            print(redbg("{} NOT defined".format(var)))

    if validate_only:
        return True

    enrich_customers = os.environ['ENRICH_CUSTOMERS']
    enrich_data = os.environ['ENRICH_DATA']

    if not validate_only:
        print("\n")
        print("Checking Data Configuration")
        print("===============\n")

    valid = os.path.exists(enrich_customers)
    if not valid:
        print(redbg("ENRICH_CUSTOMERS is missing"))

    if valid:
        customers = os.listdir(enrich_customers)
        if len(customers) == 0:
            print(redbg("No applications linked in ENRICH_CUSTOMERS {}".format(enrich_customers)))
        else:
            for c in customers:
                for usecase in os.listdir(os.path.join(enrich_customers, c)):
                    usecase_root = os.path.join(enrich_customers, c, usecase)
                    data_root = os.path.join(enrich_data, c, usecase)
                    if not (os.path.isdir(usecase_root) and
                            os.path.exists(os.path.join(usecase_root,
                                                        'enrich.json'))):
                        continue
                    if not os.path.exists(data_root):
                        print(redbg("data_root not created: {}".format(data_root)))
                    else:
                        print(greenbg("[{}] {} data_root exists".format(c, usecase)))


    siteconf = os.path.join(os.environ['ENRICH_ETC'],'siteconf.json')
    if isinstance(context, dict) and ('siteconf' in context):
        siteconf = context['siteconf']
    if not os.path.exists(siteconf):
        print(redbg("siteconf missing: {}".format(siteconf)))
    else:
        try:
            json.load(open(siteconf))
            print(greenbg("Valid siteconf exists".format(siteconf)))
        except:
            print(redbg("siteconf exists but invalid"))


    versionmap = os.path.join(os.environ['ENRICH_ETC'],'versionmap.json')
    if isinstance(context, dict) and ('versionmap' in context):
        versionmap = context['versionmap']
    if not os.path.exists(versionmap):
        print(redbg("versionmap missing: {}".format(versionmap)))
    else:
        try:
            json.load(open(versionmap))
            print(greenbg("Valid versionmap exists".format(siteconf)))
        except:
            print(redbg("versionmap exists but invalid"))

    return True

def print_log(logfile):
    """
    Show logfile
    """

    if not logfile.endswith(".json"):
        print("Requires a json file")
        return

    lines = open(logfile).readlines()
    lines = [l.strip() for l in lines]

    for l in lines:
        try:
            d = json.loads(l)
            print(" - " .join([d[col] for col in ['asctime', 'application', 'transform', 'message'] if col in d]))
            if (('data' in d) and (len(d['data']) > 0)):
                data = d['data']
                data = data.split("\n")
                data = ["     " + x for x in data]
                data = "\n".join(data)
                print("data:", data)
        except:
            print("Cannot parse ---")
            print(l)


######################################################
# Package testing...
######################################################

def print_exc(params):

    pkgdir = params['pkgdir']
    formatted_lines = traceback.format_exc().splitlines()
    exists = any([pkgdir in l for l in formatted_lines])
    skip = True
    for l in formatted_lines:
        if pkgdir in l:
            skip = False
        if not skip or not exists:
            print(l)

####################################
# Pytest
####################################
def test_structure(params):

    assert 'modname' in params
    assert 'pkgdir'in params

    modname = params['modname']
    pkgdir = params['pkgdir']
    testdir = os.path.join(pkgdir, 'tests')
    if not os.path.exists(testdir):
        print(redbg("Test directory missing"))

    return True

####################################
# Pytest
####################################
def test_pytest(params):

    assert 'modname' in params
    assert 'pkgdir'in params

    nextstep = False
    pkgdir = params['pkgdir']

    nextstep = True
    import pytest
    exitcode = pytest.main([pkgdir])
    meaning = {
        0: "All tests were collected and passed successfully",
        1: "Tests were collected and run but some of the tests failed",
        2: "Test execution was interrupted by the user",
        3: "Internal error happened while executing tests",
        4: "pytest command line usage error",
        5: "No tests were collected",
    }

    if exitcode != 0:
        if exitcode in meaning:
            print(redbg(meaning[exitcode]))
        else:
            print(redbg("Pytest failed with unknown exit code: {}".format(exitcode)))
        nextstep = False
    else:
        print(greenbg('Completed successfully'))

    return nextstep

####################################
# Loading
####################################
def test_loading(params):

    assert 'modname' in params
    assert 'pkgdir'in params

    nextstep = False
    try:
        modname = params['modname']
        pkgdir  = params['pkgdir']
        if params.get('isfile', False):
            file_, path_, desc_ = imp.find_module(modname, [pkgdir])
            params['mod'] = imp.load_module(modname, file_, path_, desc_)
        else:
            cmd = """import {}""".format(modname)
            exec(cmd)

        if modname not in sys.modules:
            print(redbg('Imported {} but cannot find in sys.modules'.format(modname)))
        else:
            params['mod'] = sys.modules[modname]
            print(greenbg('Loaded imported the {} module'.format(modname)))
            nextstep = True
    except:
        print(redbg("Unable to import {}".format(modname)))
        print_exc(params)

    return nextstep

####################################
# Instantiation
####################################
def test_instantiation(params):

    nextstep = False
    if 'mod' not in params or params['mod'] is None:
        print(redbg("Cannot execute instantiation test. Missing module"))
        return nextstep

    mod = params['mod']
    if not hasattr(mod, 'provider'):
        print(redbg("Does not have a provider attribute"))
    else:
        print(greenbg('Module has a provider attribute'))
        nextstep = True

    if not nextstep:
        return nextstep

    provider = mod.provider
    try:
        config = params['config']
        t = provider(config=config)
        params['obj'] = t
        nextstep = True
        print(greenbg('Able to instantiate the module'))
    except:
        print(redbg("Cannot instantiate the module"))
        print_exc(params)
        nextstep = False

    return nextstep

####################################
# Create mock state
####################################
def test_create_state(params):

    nextstep = False
    obj = params['obj']
    config = params['config']

    nextstep = False
    if not hasattr(obj, 'testdata'):
        print(redbg('Module does not have or has invalid testdata to test execution'))
    else:
        print(greenbg('Module has testdata'))
        nextstep = True

    try:
        # If it is a lazy function, then call it now...
        if nextstep and callable(obj.testdata):
            obj.testdata = obj.testdata()
    except:
        print(redbg('Module could not load test data'))
        nextstep = False

    if not nextstep:
        return nextstep

    if hasattr(obj, 'testdata'):
        if (('data' not in obj.testdata) or
            (not isinstance(obj.testdata['data'], dict))):
            print(redbg("Module's testdata should have dict 'data' element"))
            nextstep = False

        if nextstep and len(obj.testdata['data']) == 0:
            print(orangebg("Module's testdata usually has non-trivial 'data' element"))

        if nextstep and obj.is_skin():
            if (('runs' not in obj.testdata) or
                (len(obj.testdata['runs']) == 0)):
                print(redbg("Renderer module's testdata should have 'runs' element"))
                nextstep = False

    if not nextstep:
        return nextstep

    try:
        obj.validate('testdata', None)
        print(greenbg('Testdata appears valid'))
    except:
        print(redbg('Testdata is invalid'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    if not hasattr(obj, 'testdata'):
        raise Exception("Testdata attribute is missing")

    try:
        config.load_test_state(obj.testdata)
        print(greenbg('Able to load test data'))
    except:
        print(redbg('Module test data could not be loaded'))
        print_exc(params)
        nextstep = False

    return nextstep

####################################
# Execution
####################################
def test_execution_compute(params):

    obj = params['obj']
    config = params['config']
    state = config.get_state()

    nextstep = True

    try:
        conf = obj.get_test_conf()
        conf['test'] = True
        obj.configure(conf)
        print(greenbg('Configured the module'))
    except:
        print(redbg('Could not configure the module'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    try:
        obj.validate('conf', None)
        obj.validate('args', None)
        print(greenbg('Validated the configuration'))
    except:
        print(redbg('Could not validate the configuration'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    try:
        obj.validate('input', None)
    except:
        print(redbg('Could not validate input'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    # Run the computation
    try:
        print(greenbg('Starting process'))
        obj.process(state)
        print(greenbg('Executed the process function'))
    except:
        print(redbg('Could not execute process'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    # => Validate the results
    try:
        obj.validate('results', state)
        print(greenbg('Validated the results'))
    except:
        print(redbg('Invalid results'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    # => Store the results
    try:
        tmpdir = config.store_test_state(obj)
        print(greenbg('Stored the results'))
        print("Results in {}".format(tmpdir))
    except:
        print(redbg('Unable to store state'))
        print_exc(params)
        nextstep = False

    return nextstep


def test_execution(params):

    obj = params['obj']
    config = params['config']
    state = config.get_state()

    if (obj.is_transform() or obj.is_model()):
        return test_execution_compute(params)

    raise Exception("Unsupported module")

###############################################
# test_conf
###############################################
def load_single_transform(path, config):
    # => Now process
    modname = os.path.basename(path)
    dirname = os.path.dirname(path)

    if os.path.isfile(path):
        modname = modname.replace(".py","")

    try:
        file_, path_, desc_ = imp.find_module(modname, [dirname])
        package = imp.load_module(modname, file_, path_, desc_)
    except:
        traceback.print_exc()
        raise Exception("Couldnt load transform: {}".format(modname))

    # => Load transform class..
    if not hasattr(package, 'provider'):
        raise Exception("Transform has no provider")

    provider = package.provider

    try:
        p = provider(config=config)
    except:
        raise Exception("Could not instantiate transform")

    return p

def test_conf_pipeline(conf, context):
    """
    Test whether modules are present...
    """

    # => Collect all paths...
    params = {
        'enrich_run_root': os.environ['ENRICH_RUN_ROOT'],
        'enrich_root': os.environ['ENRICH_ROOT'],
        'enrich_customers_dir': os.environ['ENRICH_CUSTOMERS'],
        'enrich_data_dir': os.environ['ENRICH_DATA']
    }

    usecase_root = conf.get('usecase_root', conf.get('customer_root',"Unknown"))
    params.update({
        'usecase_root': usecase_root % params,
        # Backward compatability
        'customer_root': usecase_root % params,
        'data_root': conf['data_root'] % params,
    })


    for col in ['data_root', 'customer_root', 'usecase_root']:
        if not os.path.exists(params[col]):
            print(redbg("{} is missing".format(col)))

    config = MockConfig(context)
    config.load_usecase(conf['conf']) # Pass path
    config.load_test_state({})

    # Prepare the sys path
    available = []

    packages = conf['paths']['packages']
    for pkg in packages:
        pkg = pkg % params
        transformdir = os.path.join(pkg, 'transforms')
        if not os.path.exists(transformdir):
            continue
        for tname in os.listdir(transformdir):
            if tname.endswith("~") or tname == "__pycache__":
                continue
            fullpath = os.path.join(transformdir, tname, tname)
            try:
                t = load_single_transform(fullpath, config)
            except Exception as e:
                print(redbg("Invalid transform: {} ({})".format(str(e), fullpath)))
                continue
            print(greenbg("Loaded transform: {} => {}".format(tname, t.name)))
            available.append(t)

    libraries = conf['paths']['libraries']
    contribpath = os.path.join(os.path.dirname(__file__),
                               "..", "contrib", "transforms")
    libraries.append(os.path.abspath(contribpath))
    for libdir in libraries:
        libdir = libdir % params
        if not os.path.exists(libdir):
            print(redbg("Library path {} is missing".format(libdir)))
            continue
        for tname in os.listdir(libdir):
            if tname.endswith("~") or tname in ["__pycache__", '__init__.py']:
                continue
            for fullpath in  [
                    os.path.join(libdir, tname, 'src', tname),
                    os.path.join(libdir, tname, tname),
                    os.path.join(libdir, tname),
            ]:
                if os.path.exists(fullpath):
                    break

            try:
                t = load_single_transform(fullpath, config)
            except Exception as e:
                print(redbg("Invalid transform: {} ({})".format(str(e), fullpath)))
                continue
            print(greenbg("Loaded transform: {} => {}".format(tname, t.name)))
            available.append(t)

    print("\n\nLooking for required Transforms:")
    for reqt in conf['transforms']['enabled']:
        name = reqt['transform']
        args = reqt['args']

        status = 'missing'
        for avt in available:
            if avt.name == name:
                print(greenbg("Found {}".format(name)))
                status = 'notconfigured'

            if status == 'missing':
                continue

            try:
                avt.preload_clean_args(args)
                status = 'configured'
            except Exception as e:
                print(redbg("Could not configure {}".format(name)))
                traceback.print_exc()
                break

            print(greenbg("Configured {}".format(name)))
            break

        if status == 'missing':
            print(redbg("Could not find {}".format(name)))

def test_task_pipeline(conf, context):
    """
    Test whether modules are present...
    """

    # => Collect all paths...
    params = {
        'enrich_customers_dir': os.environ['ENRICH_CUSTOMERS'],
        'enrich_data_dir': os.environ['ENRICH_DATA']
    }

    params.update({
        'customer_root': conf['customer_root'] % params,
        'data_root': conf['data_root'] % params,
    })

    config = MockConfig(context)
    config.load_usecase(conf['conf'])
    config.load_test_state({})

    libraries = conf['paths']['libraries']
    for l in libraries:
        l = l % params
        if not os.path.exists(l):
            continue
        sys.path.append(l)
        for d in os.listdir(l):
            try:
                impcmd = "from {} import *".format(d)
                exec(impcmd)
                print(greenbg("Loaded: {} from {}".format(d,l)))
            except:
                print(redbg("Could not load: {}".format(d)))

    def all_subclasses(cls):
        return cls.__subclasses__() + \
            [g for s in cls.__subclasses__() for g in all_subclasses(s)]

    tasks = [c.NAME for c in all_subclasses(Task)]

    print("\n\nKnown Tasks:")
    for t in tasks:
        print(greenbg("Loaded: " + t))

    print("\n\nRequired Tasks:")
    # Known tasks
    for t in conf['tasks']:
        name = t['name']
        if name not in tasks:
            print(redbg("Could not find {}".format(name)))
        else:
            print(greenbg("Found {}".format(name)))

def test_conf(conf, capture=False, context=None):
    """
    Unit test the pipeline configuration file

    Parameters
    ----------
    conf: configuration file

    Yields
    ------
    Success/failure for each of multiple sanity checks

    """

    context = Context(context).asdict()

    # Setup logging
    if capture:
        setup_logging()

    # => Add libraries from existing deployments
    load_customer_assets()

    if not conf.lower().endswith(".json") and not conf.lower().endswith(".py"):
        print(redbg("configuration file should end with .json or .py"))
        return
    else:
        print(greenbg("Valid suffix"))

    if not os.path.exists(conf):
        print(redbg("Configuration path doesnt exist: {}".format(conf)))
        return
    else:
        print(greenbg("Valid path"))

    try:
        if conf.lower().endswith(".json"):
            confobj = json.load(open(conf))
            confobj['conf'] = conf
        else:
            # Load python
            modname = "".join([x if x.isalnum() else "_" for x in conf])
            modspec = importlib.util.spec_from_file_location(modname, conf)
            mod = importlib.util.module_from_spec(modspec)
            modspec.loader.exec_module(mod)
            if not hasattr(mod, 'config'):
                print(redbg("Module missing config element: {}".format(conf)))
                return
            confobj = mod.config
            confobj['conf'] = conf
    except:
        traceback.print_exc()
        print(redbg("Invalid configuration file"))
        return

    print(greenbg("Valid configuration file"))

    if 'transforms' in confobj:
        test_conf_pipeline(confobj, context)
    else:
        test_task_pipeline(confobj, context)

###############################################
# test_transform
###############################################
def test_transform(pkgdir, capture=False, context=None):
    """
    Unit test the module (transforms and models for now)

    Recursively test all the packages within the specified target
    directory. This will test loading, instantiation, whether state
    being updated and execution.


    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    Success/failure for each of multiple sanity checks

    """

    pkgdir = os.path.abspath(pkgdir)

    # Setup logging
    if capture:
        setup_logging()

    # => Add libraries from existing deployments
    load_customer_assets(context)

    # Cleanup pkgdir
    pkgdir = pkgdir.strip()
    if pkgdir.endswith("/"):
        pkgdir = pkgdir[:-1]

    if not os.path.exists(pkgdir):
        print("Invalid directory: {}".format(pkgdir))
        return

    config = MockConfig(context)

    # pkgdir could be
    #   x.py
    #   x/__init__.py
    #   x/x/__init__.py
    #   x/src/x/__init__.py
    #

    modname = None
    if os.path.isfile(pkgdir):
        dirname = os.path.dirname(pkgdir)
        modname = os.path.basename(pkgdir)
        modname = modname.split(".")[0]
    else:
        if not os.path.exists(os.path.join(pkgdir, 'setup.py')):
            dirname = os.path.dirname(pkgdir)
            modname = os.path.basename(pkgdir)
        else:
            modname = os.path.basename(pkgdir)
            if os.path.exists(os.path.join(pkgdir, modname)):
                dirname = pkgdir
            elif os.path.exists(os.path.join(pkgdir, 'src', modname)):
                dirname = os.path.join(pkgdir, 'src')
            else:
                raise Exception("Unable to find module")


    sys.path.append(dirname)
    if modname is None:
        raise Exception("Could not load the module")

    # Load the enrich json available...
    try:
        config.load_usecase(pkgdir)
    except:
        raise Exception("Unable to load usecase details")

    params = {
        'path': pkgdir,
        'isfile': os.path.isfile(pkgdir),
        'pkgdir': dirname,
        'modname': modname,
        'config': config,
    }

    tests = [
        #test_structure,
        # test_pytest,
        test_loading,
        test_instantiation,
        test_create_state,
        test_execution,
    ]
    for i, t in enumerate(tests):
        nextstep = t(params)
        if not nextstep:
            if i < len(tests) - 1:
                print("Skipping the rest of the testing")
            break

###############################################
# testdata list
###############################################
def load_transform(pkgdir, capture, context):

    if capture:
        setup_logging()

    # => Add libraries from existing deployments
    load_customer_assets()

    # Cleanup pkgdir
    pkgdir = pkgdir.strip()
    if pkgdir.endswith("/"):
        pkgdir = pkgdir[:-1]

    if not os.path.exists(pkgdir):
        print("Invalid directory: {}".format(pkgdir))
        return None

    # Add it to the path..
    sys.path.append(pkgdir)


    # Load the module...
    found = False
    for dirname, subdirs, files in os.walk(pkgdir):

        if 'setup.py' not in files:
            continue

        found = True
        print("\nChecking: {}".format(dirname))
        break

    if not found:
        print(redbg('Not a valid transform package'))
        return None

    # Create a config and load customer
    config = MockConfig(context)
    config.load_usecase(dirname)

    params = {
        'pkgdir': dirname,
        'modname': os.path.basename(dirname),
        'config': config,
    }

    tests = [
        #test_structure,
        # test_pytest,
        test_loading,
        test_instantiation,
        test_create_state,
    ]

    failure = False
    for i, t in enumerate(tests):
        nextstep = t(params)
        failure = not nextstep
        if failure:
            if i < len(tests) - 1:
                print("Skipping the rest of the testing")
                break

    if failure:
        return

    obj = params['obj']
    state = config.get_state()

    try:
        conf = obj.get_test_conf()
        conf['test'] = True
        obj.configure(conf)
        print(greenbg('Configured the module'))
    except:
        print(redbg('Could not configure the module'))
        print_exc(params)

    return params

def load_spec(specfile, capture):

    if capture:
        setup_logging()

    # => Add libraries from existing deployments
    load_customer_assets()

    if not os.path.exists(specfile):
        print(redbg("Invalid file: {}".format(specfile)))
        return None

    # Load python
    try:
        specfile = os.path.abspath(specfile)
        modname = "".join([x if x.isalnum() else "_" for x in specfile])
        modspec = importlib.util.spec_from_file_location(modname, specfile)
        mod = importlib.util.module_from_spec(modspec)
        modspec.loader.exec_module(mod)
        if not hasattr(mod, 'registry'):
            print(redbg("Module missing datasets element: {}".format(specfile)))
            return
    except:
        traceback.print_exc()
        print(redbg("Unable to load module"))
        return

    return mod.registry

def testdata_transform_list(pkgdir, capture, context):
    """
    List the available datasets to process

    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    A list of available datasets

    """

    # Load the transform...
    params = load_transform(pkgdir, capture, context)
    if params is None:
        print(redbg("Could not load the transform"))
        return

    # now check the testdata
    transform = params['obj']
    if 'datasets' not in transform.testdata:
        print(redbg("No datasets specified in testdata"))
        return

    datasets = transform.testdata['datasets']
    available = datasets.get('available', [])
    for i, d in enumerate(available):
        print("[{:2}] {}".format(i, d))

def testdata_spec_list(specfile, capture, context):
    """
    List the available datasets to process

    Parameters
    ----------
    spec: specfile

    Yields
    ------
    A list of available datasets

    """

    # Load the transform...
    registry = load_spec(specfile, capture)
    if registry is None:
        print("Registry not found")
        return

    for i, d in enumerate(registry.list()):
        print("[{:2}] {}".format(i, d))

def testdata_transform_show(pkgdir, capture, context, name):
    """
    Show details of a dataset

    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    A list of available datasets

    """

    # Load the transform...
    params = load_transform(pkgdir, capture, context)
    if params is None:
        print(redbg("Could not load the transform"))
        return

    transform = params['obj']
    config    = params['config']

    if 'datasets' not in transform.testdata:
        print(redbg("No datasets specified in testdata"))
        return

    datasets = transform.testdata['datasets']
    datasets = transform.testdata['datasets']
    available = datasets.get('available', [])
    extra = datasets.get('params', {})

    found = False
    for d in available:
        if d.name == name:
            found = True
            break

    if not found:
        print(redbg("Dataset {} not found".format(name)))
        return

    print(greenbg("Found dataset - {}".format(name)))


    print("Description:", d.description)

    print("------")
    print("[Targets] Available")
    print("------")

    for p in d.get_paths():
        name = p['name']
        path = p['path']
        print("  [{}] {}".format(name, path))

    print("[match] Pattern:", d.match['pattern'])

    dt = datetime.now().replace(hour=0, minute=0,
                                second=0,microsecond=0)
    examples = d.generate(end=dt,
                          start=dt + timedelta(days=-7))

    for example in examples:
        print("   :", example['timestamp'], "=>", example['name'])

def testdata_spec_show_file(config, registry, dataset):
    """
    Show commands involving a file
    """

    extra = registry.params

    print("[Targets] Available ", end='')
    print("(files)" if dataset.isfile else "(dirs)")
    print("------")
    for p in dataset.get_paths():
        name = p['name']
        path = p['path']
        nature = p['nature']
        resolved = config.get_file(path,
                                   extra=extra,
                                   abspath=False)
        print("  [{}] (type: {}) {}".format(name, nature, resolved))

    print("------")
    print("[match] Pattern:", dataset.match['pattern'])
    print("------")
    dt = datetime.now().replace(hour=0, minute=0,
                                second=0,microsecond=0)
    examples = dataset.generate(end=dt,
                                start=dt + timedelta(days=-7))

    for example in examples:
        print("   :", example['timestamp'], "=>", example['name'])

    print("------")
    print("Supported Commands")
    print("------")
    targets = defaultdict(list)
    for p in dataset.get_paths():
        name = p['name']
        nature = p['nature']
        targets[nature].append(name)

    for c in registry.get_commands(source_type="File"):
        name   = c['name']
        desc   = c['description']
        args   = c['args']

        found = True
        for arg in args:
            values = arg.find(dataset)
            if ((values is None) or
                (values in ['']) or
                ((isinstance(values, list)) and (len(values) == 0))):
                found = False
        if not found:
            continue

        print("\n[{}] {}".format(name, desc))
        example = ""
        for arg in args:
            values = arg.find(dataset)

            if not isinstance(values, list):
                values = [str(values)]
            else:
                values = [str(v) for v in values]
            print("   {:10s}:".format(arg.name),
                  ",".join(values))

            example += "{}='{}' ".format(arg.name, values[0])
        print("\n   Example: ...",name, example)

def testdata_spec_show_database(config, registry, dataset):
    """
    Show commands involving a database
    """
    for c in registry.get_commands(source_type="DBTable"):
        pass

def testdata_spec_show(specfile, capture, context, name):
    """
    Show details of a dataset

    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    A list of available datasets

    """

    context = Context(context).asdict()
    config = MockConfig(context)

    # Load the transform...
    registry = load_spec(specfile, capture)
    if registry is None:
        print(redbg("Registry is None"))
        return

    print(greenbg("Found registry"))

    extra = registry.params

    dataset = registry.find(name)
    if dataset is None:
        print(redbg("Dataset {} not found".format(name)))
        return

    print(greenbg("Found dataset - {}".format(name)))

    extra.pop('enrich_data_dir',None)
    extra.update({
        'enrich_data_dir': os.path.expandvars("$ENRICH_DATA"),
    })

    print("Description:", dataset.description)
    print("------")
    if dataset._type == 'File':
        testdata_spec_show_file(config, registry, dataset)


def testdata_transform_action(pkgdir, capture, context,
                              name, target, start, end):
    """
    Show details of a dataset

    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    A list of available datasets

    """

    # Load the transform...
    params = load_transform(pkgdir, capture, context)
    if params is None:
        print(redbg("Could not load the transform"))
        return

    transform = params['obj']
    config    = params['config']

    if 'datasets' not in transform.testdata:
        print(redbg("No datasets specified in testdata"))
        return

    datasets  = transform.testdata['datasets']
    available = datasets.get('available', [])
    extra     = datasets.get('params', {})
    command   = datasets.get('command',
                             "[%(timestamp)s]: %(backuppath)s => %(targetpath)s")

    found = False
    for d in available:
        if d.name == name:
            found = True
            break

    if not found:
        print(redbg("Dataset {} not found".format(name)))
        return

    print(greenbg("Found dataset - {}".format(name)))


    print("------")
    print("Description:", d.description)
    print("------")

    testpath = config.get_file(d.test, extra=extra, abspath=False)
    print("[test]", testpath)

    localpath = config.get_file(d.local, extra=extra)
    print("[local]", localpath)

    backuppath = config.get_file(d.backup, extra=extra, abspath=False)
    print("[backup]", backuppath)
    print("[match] Pattern:", d.match['pattern'])

    start = dateparser.parse(start) # naive datetime
    end = dateparser.parse(end)

    combinations = d.generate(end=end,start=start)
    for c in combinations:
        params = copy.copy(extra)
        testpath = os.path.join(testpath, c['name'])
        localpath = os.path.join(localpath, c['name'])
        targetpath = testpath if target == 'test' else localpath
        params.update({
            'targetpath': targetpath,
            "backuppath": os.path.join(backuppath, c['name'])
        })
        print(command % params)

def testdata_spec_action(specfile, capture, context, name,
                         command, **kwargs):
    """
    Show details of a dataset

    Parameters
    ----------
    pkgdir: Directory where the modules are.

    Yields
    ------
    A list of available datasets

    """

    context = Context(context).asdict()
    config = MockConfig(context)

    # Load the transform...
    registry = load_spec(specfile, capture)
    if registry is None:
        print(redbg("Registry is None"))
        return

    print(greenbg("Found registry"))

    extra = registry.params

    dataset = registry.find(name)
    if dataset is None:
        print(redbg("Dataset {} not found".format(name)))
        return

    print(greenbg("Found dataset - {}".format(name)))

    extra.pop('enrich_data_dir',None)
    extra.update({
        'enrich_data_dir': os.path.expandvars("$ENRICH_DATA"),
    })


    print("------")
    print("Description:", dataset.description)
    print("------")

    try:
        command = registry.get_command(command)
    except:
        print(redbg("Command {} could not be found".format(command)))
        return

    print("Command")
    print("   ", command['command'])
    print("------")

    resolve = {}
    args = command['args']
    for arg in args:
        name = arg.name
        if arg.name not in kwargs:
            raise Exception("Missing argument '{}' ({})".format(arg.name, arg.description))

        value = kwargs[arg.name]
        if not arg.validate(dataset,value):
            raise Exception("Invalid argument '{}' ({}) has value '{}'".format(arg.name, arg.description, value))

        try:
            resolve[arg.name] = arg.resolve(dataset, value,resolve=registry.params)
        except:
            raise Exception("Invalid argument '{}' ({}) has value '{}'".format(arg.name, arg.description, value))

    if (('start' not in resolve) or
        ('end' not in resolve)):
        raise Exception("'start' and 'end' datetime values are required")

    print("Args")
    print("------")
    for a in args:
        print("   ", a.name, ":", resolve[a.name])
    print("")

    print("[commands] Please run these after checking")
    combinations = dataset.generate(end=resolve['end'],start=resolve['start'])
    for c in combinations:
        params = copy.copy(registry.params)
        params['source']      = resolve['src']
        params['destination'] = resolve['dst']
        params['subdir']      = c['name']
        params['slash']       = "" if dataset.isfile else "/"

        try:
            print(command['command'] % params)
        except Exception as e:
            print(redbg("Could not generate command"))
            print(redbg("Please check registry params"))
            print("Error:", str(e))
            return

def testdata_action(action, *args, **kwargs):

    spec = args[0]

    if os.path.isdir(spec):
        print(redbg("Deprecated. Use the dataset file"))
        return

        if action == 'list':
            testdata_transform_list(*args, **kwargs)
        elif action == 'show':
            testdata_transform_show(*args, **kwargs)
        elif action == 'action':
            testdata_transform_action(*args, **kwargs)
    else:
        if action == 'list':
            testdata_spec_list(*args, **kwargs)
        elif action == 'show':
            testdata_spec_show(*args, **kwargs)
        elif action == 'action':
            testdata_spec_action(*args, **kwargs)

###############################################
# Main test
###############################################
def test_task_instantiation(params):

    nextstep = False
    if 'mod' not in params or params['mod'] is None:
        print(redbg("Cannot execute instantiation test. Missing module"))
        return nextstep

    mod = params['mod']
    clsmembers = inspect.getmembers(mod, inspect.isclass)
    #[('OrderedDict', <class 'collections.OrderedDict'>), ('StoreMetadataTask', <class 'store_metadata.StoreMetadataTask'>), ('Task', <class 'enrichsdk.tasks.Task'>)]
    # [('AnalyzeS3SourceBaseTask', <class 'enrichapp.core.tasks.analyzesource.AnalyzeS3SourceBaseTask'>), ('AnalyzeS3SourceDefaultTask', <class 'analyzesource.AnalyzeS3SourceDefaultTask'>)]

    clsmembers = [c for c in clsmembers if issubclass(c[1], Task) and c[1] is not Task ]

    if len(clsmembers) == 0:
        print(redbg("Does not have any task classes"))
    else:
        print(greenbg('Module has task classes'))
        nextstep = True

    if not nextstep:
        return nextstep

    if len(clsmembers) > 1:
        print(orangebg("Module has more than task. Testing only one"))
        provider = None
        modpath = mod.__file__
        for c in clsmembers:
            path = inspect.getfile(c[1])
            if path == modpath:
                provider = c[1]

        if provider is None:
            # Could not locate the right task to test
            print(orangebg(str(clsmembers)))
            print(orangebg(str("Mod path", modpath)))
            raise Exception("Could not find right task provider class to test")
    else:
        provider = clsmembers[0][1]

    try:
        config = params['config']
        t = provider(config=config)
        params['obj'] = t
        nextstep = True
        print(greenbg('Able to instantiate the module'))
    except:
        print(redbg("Cannot instantiate the module"))
        print_exc(params)
        nextstep = False

    return nextstep

def test_task_create_state(params):

    nextstep = False
    obj = params['obj']
    config = params['config']

    nextstep = False
    if not hasattr(obj, 'testdata'):
        print(redbg('Module does not have or has invalid testdata to test execution'))
    else:
        print(greenbg('Module has testdata'))
        nextstep = True


    if hasattr(obj, 'testdata'):
        # If callable, then call it...
        try:
            if callable(obj.testdata):
                obj.testdata = obj.testdata()
        except:
            print(redbg('Module could not load test data'))
            nextstep = False

        if not nextstep:
            return nextstep

        if (('data' not in obj.testdata) or
            (not isinstance(obj.testdata['data'], dict))):
            print(redbg("Module's testdata should have dict 'data' element"))
            nextstep = False

        if nextstep and len(obj.testdata['data']) == 0:
            print(orangebg("Module's testdata usually has non-trivial 'data' element"))

    if not nextstep:
        return nextstep

    try:
        obj.validate('testdata', None)
        print(greenbg('Testdata appears valid'))
    except:
        print(redbg('Testdata is invalid'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    try:
        config.load_test_state(obj.testdata)
        print(greenbg('Able to load test data'))
    except:
        print(redbg('Module test data could not be loaded'))
        print_exc(params)
        nextstep = False

    return nextstep

def test_task_load_testdata(params):

    nextstep = False
    obj = params['obj']
    config = params['config']

    nextstep = False
    if not hasattr(obj, 'testdata'):
        print(redbg('Module does not have or has invalid testdata to test execution'))
    else:
        print(greenbg('Module has testdata'))
        nextstep = True

    if hasattr(obj, 'testdata'):
        if (('data' not in obj.testdata) or
            (not isinstance(obj.testdata['data'], dict))):
            print(redbg("Module's testdata should have dict 'data' element"))
            nextstep = False

        if nextstep and len(obj.testdata['data']) == 0:
            print(orangebg("Module's testdata usually has non-trivial 'data' element"))

    if not nextstep:
        return nextstep

    try:
        obj.validate('testdata', None)
        print(greenbg('Testdata appears valid'))
    except:
        print(redbg('Testdata is invalid'))
        print_exc(params)
        nextstep = False

    return nextstep

def test_task_execution_compute(params):

    obj = params['obj']
    config = params['config']
    state = config.get_state()

    nextstep = True

    try:
        config_params = obj.testdata.get('config', {})
        config.configure(config_params)
    except:
        print(redbg('Could not configure the mock pipeline executor'))
        print_exc(params)
        nextstep = False

    try:
        conf = obj.testdata.get('conf', {})
        obj.configure(conf)
        print(greenbg('Configured the module'))
    except:
        print(redbg('Could not configure the module'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    try:
        obj.validate('conf', state)
        obj.validate('args', state)
        print(greenbg('Validated the configuration'))
    except:
        print(redbg('Could not validate the configuration'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    # Run the computation
    try:
        print(greenbg('Starting task execution'))
        obj.run(state)
        print(greenbg('Executed the process function'))
    except:
        print(redbg('Could not execute process'))
        print_exc(params)
        nextstep = False

    if not nextstep:
        return nextstep

    # => Validate the results
    try:
        obj.validate('results', state)
        print(greenbg('Validated the results'))
    except:
        print(redbg('Invalid results'))
        print_exc(params)
        nextstep = False

    print("State")
    print(state.state)

    return nextstep


def test_task(taskdir, capture=False, context=None):
    """
    Unit test the task module

    Parameters
    ----------
    tasklib: Directory where the task lib is

    Yields
    ------
    Success/failure for each of multiple sanity checks
    """

    context = Context(context).asdict()

    if capture:
        setup_logging()

    # => Add libraries from existing deployments
    load_customer_assets()

    # Cleanup pkgdir
    taskdir = taskdir.strip()
    if taskdir.endswith("/"):
        taskdir = taskdir[:-1]

    if not os.path.exists(taskdir) or os.path.isfile(taskdir):
        print("Invalid directory: {}".format(taskdir))
        return

    # Add it to the path..
    dirname = os.path.dirname(taskdir)
    sys.path.append(dirname)

    params = {
        'pkgdir': dirname,
        'modname': os.path.basename(taskdir),
        'config': MockConfig(context)
    }

    tests = [
        test_loading,
        test_task_instantiation,
        test_task_create_state,
        test_task_load_testdata,
        test_task_execution_compute
    ]

    for i, t in enumerate(tests):
        nextstep = t(params)
        if not nextstep:
            if i < len(tests) - 1:
                print("Skipping the rest of the testing")
                break




import os
import json
import copy
import re
import logging
import traceback
import importlib
import subprocess
from datetime import datetime
from dateutil import parser as dateparser
from dateutil.tz import tzlocal
from colored import fg, bg, attr

from .exceptions import *
from .context import *
from .resources import *

logger = logging.getLogger('app')

def greenbg(txt):
    return "{} \u2713 {} {}".format(bg('light_green'), txt, attr(0))

def redbg(txt):
    return "{} \u274C {} {}".format(bg('indian_red_1a'), txt, attr(0))

def orangebg(txt):
    return "{} \u0023 {} {}".format(bg('dark_orange_3a'), txt, attr(0))

cached_siteconf = {}
cached_versionmap = {}

def read_siteconf_json(filename, app={}, context={}):

    global cached_siteconf

    try:
        # Try the sitelib first.
        import sitelib

        # Return cached file
        if filename in cached_siteconf:
            return cached_siteconf[filename]

        nature, siteconf = sitelib.read_siteconf(filename, app)
        if siteconf is None:
            raise Exception("Failed to read siteconf")

        # Cache it
        cached_siteconf[filename] = siteconf
        if nature == 'open':
            logger.warning("Successfully read siteconf but it is insecure")
        else:
            logger.warning("Successfully read secure siteconf")
    except:
        # sitelib may not be installed
        logger.debug("Could not read secure siteconf. Falling back")
        try:
            # If this is a jsonx file, then try the json version of
            # the file.
            if filename.endswith('jsonx'):
                filename = filename.replace('jsonx', 'json')
            siteconf = json.load(open(filename))
            cached_siteconf[filename] = siteconf
        except:
            traceback.print_exc()
            siteconf = None

    if siteconf is None:
        raise Exception("Could not read siteconf")

    if not isinstance(siteconf, dict):
        raise Exception("Invalid siteconf found. Expecting 'dict' found '{}'".format(str(type(siteconf))))

    return siteconf

def read_siteconf(filename=None, app={}, context={}):


    # If context not available, then generate it.
    if len(context) == 0:
        context = Context().asdict()

    if filename is not None:
        return read_siteconf_json(filename, app)

    tries = []
    try:
        tries.extend([
            "%(ENRICH_ETC)s/siteconf.jsonx" % context,
            "%(ENRICH_ETC)s/siteconf.json" % context,
        ])
    except:
        pass

    tries.extend([
        os.path.expandvars("${ENRICH_ETC}/siteconf.jsonx"),
        os.path.expandvars("${ENRICH_ETC}/siteconf.json"),
    ])

    siteconf = None
    for filename in tries:
        if os.path.exists(filename):
            siteconf = read_siteconf_json(filename, app, context)
            if siteconf is None:
                continue
            else:
                break

    if siteconf is None:
        raise Exception("Could not find siteconf. Tried {} alternatives".format(len(tries)))

    return siteconf

def get_credentials(siteconf, name):
    """
    Look up the credentials file
    """

    if ((siteconf is None) or
        (not isinstance(siteconf, dict))):
        raise Exception("Invalid siteconf - None or has invalid datatype")

    if (('credentials' not in siteconf) or
        (name not in siteconf['credentials'])):
        raise Exception("missing credentials")

    return siteconf['credentials'][name]

def get_credentials_by_name(name, app={}, context={}):
    """
    Lookup credentials by name
    """

    siteconf = read_siteconf(app=app, context={})
    return get_credentials(siteconf, name)


def get_credentials_by_type(conditions,app={},context={}):
    """
    Lookup credentials by type
    """

    siteconf = read_siteconf(app=app, context={})

    if ((siteconf is None) or
        (not isinstance(siteconf, dict))):
        raise Exception("Invalid siteconf - None or has invalid datatype")

    if 'credentials' not in siteconf:
        raise Exception("missing credentials")

    if not isinstance(conditions, dict):
        raise Exception("Invalid condition specification")

    final = {}
    for name, detail in siteconf['credentials'].items():

        if not isinstance(detail, dict):
            continue

        if not detail.get('enable', True):
            continue

        match = True
        for var, choices in conditions.items():
            if var not in detail:
                match = False
                break
            if detail[var] not in choices:
                match = False
                break

        if not match:
            continue

        final[name] = copy.deepcopy(detail)

    return final

def update_from_git(detail, context,debug=False):

    # Make a local version
    context = copy.copy(context)

    label = detail.get('label', detail.get('repo','unknown'))

    # Try the detail run root
    root = detail['git'] % context
    if not os.path.exists(root):

        # try one alternative path if possible
        if "ENRICH_RUN_ROOT" not in detail['git']:
            raise Exception("Expected ENRICH_RUN_ROOT: {}".format(detail['git']))

        # Fallback to the default run_root
        context['ENRICH_RUN_ROOT'] = context['ENRICH_ROOT']
        altroot = detail['git'] % context
        if not os.path.exists(altroot):
            raise Exception("Missing repo roots: {}, {}".format(root, altroot))
        root = altroot

    # Update in place
    detail['git'] = root

    # By default use master branch only..
    branch = detail.get('branch', 'master')

    # Collect all the notes
    cmd = "git -C {} log {} --pretty='%h--%ci--%N'".format(root, branch)
    notes = subprocess.check_output(cmd, shell=True)
    notes = notes.decode('utf-8')
    notes = notes.split("\n")
    notes = [n.split("--") for n in notes if (n.count("--") == 2 and not n.endswith("--"))] # remove empty

    detail['notes']= [
        {
            'commit': n[0],
            'commit_date': n[1],
            'message': n[2]
        } for n in notes
    ]

    # => Collect branches..
    cmd = """git -C {} branch -l""".format(root)
    lines = subprocess.check_output(cmd, shell=True)
    lines = lines.decode('utf-8')
    lines = lines.split("\n")
    lines = [l[2:] for l in lines]
    detail['branches'] = lines

    # Collect the git commits
    cmd = """git -C {} log {} --no-walk --tags --all --pretty="%h\t%D\t%ce\t%ci\t%s" --decorate=full --date=short""".format(root, branch)
    lines = subprocess.check_output(cmd, shell=True)
    lines = lines.decode('utf-8')
    lines = lines.split("\n")
    lines = [l.strip() for l in lines]
    lines = [l.split("\t") for l in lines]
    lines = [l for l in lines if len(l) == 5]
    alltags = []
    for l in lines:

        if debug:
            print(l)

        # Skip all irrelevant remotes except upstream for the given branch
        refs = [x.strip() for x in l[1].split(",")]

        # Use relevant records. There is stuff from other branches,
        valid = False
        for r in refs:
            if (('refs/heads/{}'.format(branch) in r) or
                ('refs/remotes/origin/{}'.format(branch) in r) or
                ('refs/tags' in r) or
                ("HEAD" in r)):
                if debug:
                    print("Match", r)
                valid = True
                break

        if not valid:
            if debug:
                print("Skipping")
            continue

        # Is this head?
        head = any([ref.startswith('HEAD') for ref in refs])

        # Align the local date
        try:
            dt = dateparser.parse(l[3]).astimezone(tzlocal()).isoformat()
        except:
            dt = l[3]

        entry = {
            #['416366c', 'tag: refs/tags/v1.4.0, refs/heads/master', '2020-03-03', 'pingali@gmail.com', 'Bug']
            'commit': l[0],
            'refs':   refs,
            'author':   l[2],
            'date': dt,
            'log':   l[4]
        }

        # Extract the tag for this commit
        for ref in refs:
            match = re.search("tag: refs/tags/(\S+)", ref)
            if match is not None:
                tag = match.group(1)
                entry['tag'] = tag
                break

        if 'tag' not in entry:
            entry['tag'] = l[0]

        if debug:
            print(entry)

        alltags.append(entry)

    # Now sort by reverse commit date
    alltags = sorted(alltags,
                     key=lambda entry: entry['date'],
                     reverse=True)
    detail['alltags'] = alltags


    position='post'
    for entry in alltags:

        head = any([ref.startswith('HEAD') for ref in entry['refs']])

        # Where is this position...
        if head:
            entry['position'] = 'head'
            position = 'pre'
        else:
            entry['position'] = position

        if head:
            detail['release'] = entry['tag']
            detail['commit'] = entry['commit']
            detail['date'] = entry['date']

            # Insert the URL as well
            if 'url' in detail:
                urlparts = detail['url'].split("/")
                urlparts[-1] = detail['commit']
                url = "/".join(urlparts)
                detail['url'] = url

def update_from_package(detail, context):

    pkgname = detail['package']
    mod = importlib.import_module(pkgname)

    # v0.4.3+0.g659ce50.dirty
    if not hasattr(mod, '__version_detailed__'):
        return

    pkgdetails = mod.__version_detailed__
    release = pkgdetails['version']
    commit  = pkgdetails['full-revisionid']

    # Now overwrite the release information
    detail['release'] = release
    detail['commit'] = commit[:8]
    if 'url' in detail:
        urlparts = detail['url'].split("/")
        urlparts[-1] = detail['commit']
        url = "/".join(urlparts)
        detail['url'] = url
    else:
        detail['release'] += " (out of date; version missing)"
        logger.error("VersionMap: Cannot collect version information for {}".format(pkgname),
                     extra={
                         'data': json.dumps(detail, indent=4)
                     })

def read_versionmap(filename=None,
                    include_notes=False,
                    include_tags=False,
                    context={}):

    global cached_versionmap

    if isinstance(context, dict):
        if len(context) == 0:
            context = Context().asdict()
    elif isinstance(context, Context):
        context = context.asdict()

    default = [
        {
            "repo": "error",
            "release": "v1xxx",
            "date": "2018-10-09 12:06:40 +0530",
            "commit": "5079404f12df939c138215ed1a02e4d0d8814e00",
            "description": "Error while reading the versionmap",
            "url": "https://github.com/pingali/error/commit/5079404f12df939c138215ed1a02e4d0d8814e00",
            "label": "error",
            "notes": [
                {
                    "commit": "6271dcae",
                    "commit_date": "2018-10-09 12:06:40 +0530",
                    "notes": "v0 of the release"
                }
            ]
        }
    ]

    if filename is None:
        if 'versionmap' in context:
            filename = context['versionmap']
        elif 'ENRICH_ETC' in context:
            filename = "%(ENRICH_ETC)s/versionmap.json" % context
        else:
            filename = os.path.expandvars("$ENRICH_ETC/versionmap.json")
    now = datetime.now().replace(microsecond=0).isoformat()
    if not os.path.exists(filename):
        error = "Missing versionmap: {}".format(filename)
        logger.error(error)
        default[0]['description'] = error
        return None, default

    # => use the filename's timestamp as the default timestamp
    timestamp = datetime.fromtimestamp(os.path.getmtime(filename))
    timestamp = timestamp.isoformat()

    if ((filename in cached_versionmap) and
        (not include_notes) and
        (not include_tags)):
        return timestamp, cached_versionmap[filename]

    try:
        versionmap = json.load(open(filename))
    except:
        error = "Invalid version file. Not a valid json file: {}".format(filename)
        logger.error(error)
        default[0]['description'] = error
        return default

    # => Newer dict format of the versionmap
    if isinstance(versionmap, dict):
        timestamp = versionmap.get('timestamp', timestamp)
        versionmap = versionmap.get('versionmap', None)

    if not isinstance(versionmap, list):
        error = "Invalid version map. Not a list: {}".format(filename)
        logger.error(error)
        default[0]['description'] = error
        return timestamp, default

    # Update the release and commit information if dynamic modules are
    # specified.
    for i in range(len(versionmap)):

        label = versionmap[i].get('label', versionmap[i].get('repo','unknown'))

        if 'branch' not in versionmap[i]:
            versionmap[i]['branch'] = 'master'

        if versionmap[i].get('dynamic', False):
            try:
                if 'package' not in versionmap[i] and 'git' not in versionmap[i]:
                    versionmap[i]['release'] += " (out of date; config issue)"
                    logger.error("VersionMap: Missing package to be checked",
                                 extra={
                                     'data': json.dumps(versionmap[i], indent=4)
                                 })
                elif 'package' in versionmap[i]:
                    update_from_package(versionmap[i], context)
                elif 'git' in versionmap[i]:
                    update_from_git(versionmap[i], context)

            except:
                traceback.print_exc()
                versionmap[i]['release'] += " (out of date; import issue)"
                logger.exception("{}: Unable to import package/repository to update version".format(label),
                                 extra={
                                     'data': json.dumps(versionmap[i], indent=4)
                                 })

    # Cleanup of the output
    for i in range(len(versionmap)):

        # Drop the notes by default
        if not include_notes:
            versionmap[i].pop('notes',[])

        if not include_tags:
            versionmap[i].pop('alltags',[])

    # Cache only if the versionmap requested is standard. Special
    # versionmaps should not be cached.
    if ((not include_tags) and (not include_notes)):
        cached_versionmap[filename] = versionmap

    return timestamp, versionmap



import os 
import subprocess 
from hashlib import sha256

def get_checksum(path): 
    
    if not os.path.isfile(path) or not os.path.exists(path): 
        return None 

    checksum = subprocess.check_output('/usr/bin/sha256sum {}'.format(path),        
                               shell=True)
    #'ffd4d29ee4019150cc308e46004d2f17105899983e7fb4ca7c27ffe981fb82d4  deploy.json\n'
    checksum = checksum.decode('utf-8').strip() 
    checksum = checksum.split(" ")[0] 
    return checksum

def get_file_size(path): 
    
    if not os.path.isfile(path) or not os.path.exists(path): 
        return 0

    statinfo = os.stat(path) 
    return statinfo.st_size



"""
"""
import os
import sys
import json
import copy

from ..lib import read_siteconf 

class Service(object):
    """
    Base class for all services.
    
    Each class of services extends this base class to achieve its
    performance and other objectives. For example, `data quality
    service` (DataSourceService) that is provided by the `Contrib`
    application from `Scribble` customer.

    """
    def __init__(self, conf, *args, **kwargs):
        """
        Initialize the service manager 
        """
        if 'name' not in conf:
            raise Exception("Invalid configuration: Every service requires a name")

        self.conf = conf 
        """
        Configuration of this service. 
        
        This is interpreted by the subclass functions 
        """

        self.name = self.conf['name']
        """
        Name of the service. 
        """

        self.context = {
            'enrich_root': os.environ['ENRICH_ROOT'],
            'enrich_root_dir': os.environ['ENRICH_ROOT'],
            'enrich_data': os.environ['ENRICH_DATA'],
            'enrich_data_dir': os.environ['ENRICH_DATA'],
        }
        """
        Context for path resolution
        """

    def get_file(self, path, extra={}):
        """
        Resolve a path 
        """

        fullcontext = copy.copy(self.context)
        fullcontext.update(extra)
        return path % fullcontext 
        
    def get_credentials(self, name):
        """
        Look up the credentials file 
        """
        
        siteconf = read_siteconf()
        if siteconf is None:
            raise Exception("Could not find siteconf") 
        
        if (('credentials' not in siteconf) or 
            (name not in siteconf['credentials'])): 
            raise Exception("missing credentials") 
            
        return siteconf['credentials'][name] 




"""
Transforms
^^^^^^^^^^

This class bridges expectations library and the Enrich pipelines. 
"""

import os
import sys
import imp
import json
import logging
import pickle
import traceback

from .exceptions import *
from .expectations import *

logger = logging.getLogger('app')

class TransformExpectation(object):
    """
    Class to provide a bridging interface between light weight
    expectations implemented here and the pipeline
    """

    def __init__(self, transform,
                 mode=None, filename=None,
                 expectations=None):
        """
        Initialize the class

        Args:
           transform: Transform using this expectation
           mode: Mode of operation (validation|generation)
           filename: Path to expectations file
           expectations: Explicitly provide expectations

        Returns:
           Instantiated class

        """
        self.transform = transform
        assert mode in ['validation', 'generation', None]

        if ((filename is None) and (expectations is None)):
            raise InvalidExpectation("Missing expectations and filename")
        elif ((filename is not None) and (expectations is not None)):
            raise InvalidExpectation("Cant specify both expectations and filename")

        if filename is not None:
            filename = transform.config.get_file(filename)

        self.mode = mode
        if self.mode == 'validation':
            if filename is not None:
                if not os.path.exists(filename):
                    raise InvalidExpectation("Missing expectations file")
            elif expectations is not None:
                if not isinstance(expectations, list):
                    raise InvalidExpectation("Expectations should be a list")

        if expectations is not None:
            self.expectations = expectations
        else:
            self.expectations = []

        try:
            if ((mode == 'validation') and (filename is not None)):
                self.expectations = self.load_expectations_file(filename)
        except:
            raise InvalidExpectation("Invalid expectation file")


    def load_expectations_file(self, filename):
        """Load expectations whether specified as json, pickle, or py
        file into a instance variable.

        Args:
            filename: Name of the expectations file

        Returns:
            None:

        """
        self.filename = filename
        if filename.endswith('.json'):
            return json.load(open(filename))

        if filename.endswith('.pkl'):
            return pickle.load(open(filename, 'rb'))

        if filename.endswith(".py"):
            try:

                modname = os.path.basename(filename)
                dirname = os.path.dirname(filename)

                if modname.endswith(".py"):
                    modname = modname.replace(".py","")

                file_, path_, desc_ = imp.find_module(modname, [dirname])
                package = imp.load_module(modname, file_, path_, desc_)
                return package.expectations
            except:
                logger.exception("Skipping. Expectations file could not be read",
                                 extra=self.get_extra({
                                     'transform': self.transform,
                                     'data': "Filename: {}".format(filename)
                                 }))
                raise InvalidExpectation("Invalid specifications file: {}".format(modname))

        raise UnsupportedExpectation("Unsupported expectations specification file")



    def generate(self, df):

        """
        Render a specified template using the context

        Args:
           df: Dataframe to profiled

        Returns:
            Rendered html template that can be embedded

        """
        if self.mode != 'generate':
            raise InvalidExpectation("Mode is not generate")

        pass

    def validate(self, df, selector=None):
        """
        Run the loaded expectations

        Args:
           df: Dataframe to evaluated
           selector: Function to select the

        Returns:
           result: A list of dictionaries, each has evaluation result
        """

        if self.mode != 'validation':
            raise InvalidExpectation("Dataframe validation not enabled")

        mgr = ExpectationManagerBase()
        mgr.initialize()

        expectations = self.expectations
        if callable(selector):
            expectations = selector(expectations)

        result = mgr.validate(df, expectations)

        try:
            json.dumps(result)
        except:
            logger.exception("Error while processing expectations",
                             extra={
                                 'transform': self.transform.name
                             })
            return []

        return result



import os
from jinja2 import Environment, FileSystemLoader, PackageLoader, meta

thisdir = os.path.dirname(__file__) 
default_templatedir = os.path.abspath(os.path.join(thisdir, '..', 'templates'))

#from prompt_toolkit.token import Token
#from prompt_toolkit.styles import style_from_dict
#style = style_from_dict({
#    Token.Toolbar: '#ffffff bg:#333333',
#})


##########################
# Jinja2 helper functions...
##########################
def get_template_path(filename, templatedir=None): 
    if os.path.isabs(filename): 
        return filename 
        
    if templatedir is None: 
        templatedir = default_templatedir 

    # Check if relative path is specified, then use it. 
    if os.path.exists(filename): 
        return filename 

    return os.path.join(templatedir, filename) 

def list_templates(suffix=".node.template"): 

    mapping = {} 
    for t in os.listdir(default_templatedir): 
        if t.endswith(suffix): 
            path = os.path.join(default_templatedir, t)
            mapping[t.replace(suffix,"")] = path 
    return mapping 

def render(tpl_path, context):
    path, filename = os.path.split(tpl_path)
    return Environment(
        loader=FileSystemLoader(path or './')
    ).get_template(filename).render(context)

def get_variables(tpl_path):     
    path, filename = os.path.split(tpl_path)
    env = Environment(loader=FileSystemLoader(path or './'))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)
    return meta.find_undeclared_variables(parsed_content)

def write_rendered_file(templatefile, target, filename, params): 

    if not templatefile.startswith("/"):
        templatefile = get_template_path(templatefile) 
        
    if not os.path.exists(templatefile): 
        raise Exception("Template file doesnt exist") 

    content = render(templatefile, params) 
    try: 
        os.makedirs(target) 
    except: 
        pass 

    with open(os.path.join(target,filename), 'w') as fd: 
        fd.write(content) 
    return content 




import re
import traceback
from texttable import Texttable

def draw_table(backend, command, result, columns, extra={}):

    # Now we have to beautify
    data = result['data']

    t = Texttable(max_width=0)
    t.add_rows([['Attribute',"Value"],
                ['Name', backend.config['name']],
                ['Server', backend.config['server']],
                ['Command', command]] +
                [[k,v] for k,v in extra.items()])
    print(t.draw())

    t = Texttable(max_width=0)
    #t.set_deco(Texttable.HEADER)
    t.add_rows([columns] +
               [[d.get(c,'') for c in columns] for d in data])

    print(t.draw())

def draw(backend, command, result, columns, extra={}):
    return draw_table(backend, command, result, columns, extra)

def draw_dict(backend, command, result, keys, extra={}):

    # Now we have to beautify
    data = result['data']

    t = Texttable(max_width=0)
    t.add_rows([['Attribute',"Value"],
                ['Name', backend.config['name']],
                ['Server', backend.config['server']],
                ['Command', command]] +
                [[k,v] for k,v in extra.items()])
    print(t.draw())

    t = Texttable(max_width=0)
    #t.set_deco(Texttable.HEADER)
    t.add_rows([["Key", "Value"]] +
               [[c, data.get(c,'')] for c in keys])

    print(t.draw())

def draw_run_detail(backend, command, result, extra={}):

    # Now we have to beautify
    data = result['data']

    t = Texttable(max_width=0)
    t.add_rows([['Attribute',"Value"],
                ['Name', backend.config['name']],
                ['Server', backend.config['server']],
                ['Command', command]] +
                [[k,v] for k,v in extra.items()])
    print(t.draw())
    print("\n")

    if 'metadata' not in data:
        return

    metadata = data['metadata']

    print("RUN DETAILS")

    rows = [['Attribute','Value'],
            ['Start', metadata['start_time']],
            ['End', metadata.get('end_time','')],
            ['Command', " \n".join(metadata['cmdline'])],
            ['User', metadata['stats']['user']],
            ['Host', metadata['stats']['platform']['node']]]

    notes = metadata.get('summary',[]) +\
            metadata.get('performance_summary',[])
    if len(notes) > 0:
        rows.append(['Summary', "\n".join(notes)])

    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")


    print("OUTPUTS")

    rows = [['Frame', 'Type', 'Description', 'Size', 'Path']]
    for name, details in metadata.get('details', {}).items():
        frametype = details['frametype']
        description = ""
        path = ""
        filesize = ""
        params = details.get('params',{})
        for p in params:
            if p.get('action',None) != 'output':
                continue

            description = p.get('descriptions', p.get('description',""))
            if isinstance(description, list):
                description = [d[:30] for d in description]
                description = "\n".join(description)
            else:
                description = description[:30]
            components = p.get('components', [])
            if len(components) > 0:
                path = "..." + components[0]['filename'][-30:]
                filesize = components[0]['filesize']
            break
        rows.append([name, frametype, description, filesize, path])

    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")

    print("VERSIONS")

    rows = [['Label', 'Description','Release', 'Commit']]
    for v in metadata['versionmap']:
        repo = v['repo']
        label = v['label']
        commit = v['commit'][:8]
        description = v['description'][:20]
        release = v['release']
        if repo.startswith('scribble-'):
            continue
        rows.append([label, description, release, commit])
    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())
    print("\n")

    if 'process' in metadata:
        print("RESOURCES")
        rows = [['Attribute', 'Value']]
        for v in metadata['process']:
            if v[0] in ['VmPeak', 'VmSize', 'MemTotal', 'MemFree']:
                rows.append(v)
        t = Texttable(max_width=0)
        t.add_rows(rows)
        print(t.draw())
        print("\n")

    print("LOG")
    log = data['log']
    rows = [['Elapsed', 'Transform', 'Level', 'Message']]
    for l in log:
        try:
            rows.append([l['elapsed'],
                         l.get('transform',''),
                         l['levelname'],
                         l['message'][:30]])
        except:
            traceback.print_exc()
    t = Texttable(max_width=0)
    t.add_rows(rows)
    print(t.draw())

def draw_health(backend, command, result, extra={}):

    # Now we have to beautify
    data = result['data']

    def fixedlength(paragraph, n):
        paragraph = paragraph.strip()
        return "\n".join([paragraph[i: i + n] for i in range(0, len(paragraph), n)])

    for section in data:
        print(section['title'])
        print("-----")
        print(section['description'])
        t = Texttable(max_width=0)
        rows = [['Option','Value','Description']]
        for o in section['options']:
            value = o['value']
            value = value.replace("<br>","\n")
            value = value.replace("</th>\n","</th>")
            value = value.replace("</td>\n","</td>")
            value = re.sub("<[^>]+>",'', value)
            name = fixedlength(o['name'], 20)
            desc = fixedlength(o['description'], 20)
            rows.append([name, desc, value])
        t.add_rows(rows)
        print(t.draw())
        print("\n")



import os
import sys
import importlib
import string
import logging
import traceback

########################################################
# Helper functions 
########################################################
def extend_run_commandset(_run, config):
    """
    Load commands from the libraries specified 
    """

    debug = config.debug    
    paths = config.get_library_paths() 

    if debug:
        logging.debug("Loading run commands")
        
    for p in paths:

        # Enable import from these directories 
        p = os.path.abspath(p)
        if not os.path.exists(p):
            if debug:
                logging.debug("Unable to load path: {}".format(p))
            continue 
        
        sys.path.append(p) 

        commands = os.listdir(p)
        for c in commands:
            
            fullpath = os.path.join(p, c)

            #####################################
            # Check if I should skip it
            #####################################
            skip = False
            suffixes = ['.pyc', '.pyo','__pycache__', '#', "~"]
            for s in suffixes:
                if c.endswith(s):
                    skip = True
                    break
            if skip:
                if debug: 
                    logging.debug("Skipping: {}".format(c))
                continue
            
            try: 
                # => Get the spec...
                # Handle directories differently...
                modname = "".join([x if x in string.ascii_letters else "_" for x in c])
                if os.path.isdir(fullpath):
                    spec = importlib.util.find_spec(c)
                else:
                    spec = importlib.util.spec_from_file_location(modname,fullpath) 
                                               
                if spec is None:
                    if debug:
                        logging.debug("can't find the module")
                    continue 

                # If you chose to perform the actual import ...
                mod = importlib.util.module_from_spec(spec)                
                if mod is None and spec.loader is not None:
                    if debug:
                        logging.debug("Loaded module is None or the loader is missing: {}".format(c))
                    continue 

                spec.loader.exec_module(mod)
                if not hasattr(mod, 'entrypoint'):
                    if debug:
                        logging.debug("Loaded module does not have 'entrypoint' attribute: {}".format(c))
                    continue

                _run.add_command(mod.entrypoint)
            except:
                if debug: 
                    traceback.print_exc() 
                pass

            if debug:
                logging.debug("Loaded module: {}".format(c))

        if debug:
            logging.debug("Completed loading commands")
            
        
        



"""Enrich has a notion of a pipeline state. The State object is used
to keep track of all available data objects (dataframes, dictionaries,
html objects etc.) and deliver it to whatever transform requires it.

The combination of transform and state allows a complex pipeline to be
split into multiple stages, and independent development. Further the
transform code can be spread across multiple repositories.

The State object associates metadata with every data object, that is
dumped into the metadata.json at the end of the run.

The metadata associated with a data object includes::

  1. Name of the frame (each data object has a unique name)
  2. Category of the data object 
  3. Type of frame (e.g., pandas, html, dict) 
  4. History of access to the dataframe 
  5. Parameters that can have arbitrary informations such as
     column documentation, security, messages etc. 
  6. Stages that each data object goes through


"""

import os
import sys
import json
import logging
import platform
import getpass
import distro
from datetime import datetime

class EnrichStateBase(object):
    """
    Base class for state manager
    """
    def __init__(self, config, *args, **kwargs):
        """
        Config object is a required parameter
        """
        self.config = config
        self.start_time = self.config.now()
        self.end_time = self.config.now()
        self.name = config.name
        self.display = config.display
        self.runid = config.runid
        self.pid = os.getpid()

        # Status...
        self.histid = 0
        self.status = 'started'
        self.summary = [] # notes that must be shared.
        self.performance_summary = []
        self.frames = {}
        self.transforms = {}
        self.notification = {
            'email': []
        }

        self.expectations = []

        # Who is running this?
        user = getpass.getuser()
        cmdline_args = self.config.get_cmdline_args()
        if 'user' in cmdline_args:
            user = cmdline_args['user']

        # Background system information
        self.stats = {
            'platform': {
                'node': platform.node(),
                'os': platform.system(),
                'release': platform.release(),
                'processor': platform.processor(),
                'python': platform.python_version(),
                'distribution': distro.linux_distribution()
            },
            'user': user
        }

    def update_frame(self, name, detail, create=False):
        """
        Update the state with the details of a dataframe

        Args: 
            name (str): Name of the dataframe 
            detail (dict): Dict with a number of details 
            create (bool): Add if the frame doesnt exist.


        The detail has the elements mentioned above::

            {
               'df': df,               # data object 
               'params': params,       # extra metadata 
               'transform': self.name, # Callee 
               'history': [
                   {
                      'log': 'Passed some additional params'
                   }
               ]
            }
         
        Params could be a dict or a list of dicts. Each 
        entry has a `type` and type-specific attributes. 
        Over a period of time a nuber of different 
        attributes were added. Some of them are::
             
              1. Compute - column information and description
              2. Overrides - information/instructions from one 
                 transform to another
              3. Lineage - Input and output dependencies for 
                 better tracking

        For example:: 

            [
                {
                    # Column metadata
                    'type': 'compute',
                    'columns': columns,
                    'description': [
                        "Cars1 Dataset with no changes"
                    ]
                },
                {
                    # 'Fake' the saving 
                    'type': 'overrides',
                    'transform': 'TableSink',
                    'args': {
                        'save': False,
                        'rows': 193332
                    }
                },
                {
                    # Dependency information
                    "type": "lineage",
                    "transform": self.name,
                    "dependencies": [
                        {
                            "type": "file",
                            "nature": "output",
                            "objects": [sqloutput, output]
                        },
                        {
                            "type": "dataframe",
                            "nature": "input",
                            "objects": ["cars1", "cars2"]
                        }
                    ]
                }
    
            ]
    
        """
        pass



import os
import json 
import argparse
from pyspark.sql import SQLContext

from .spark import start_spark 

"""
This is a wrapper around start_spark function. This 
will take the configuration parameters specified in 
'config' and also enrich environment to create the 
spark session 
"""

def start_spark_session(config):
    """
    Main script

    :return: None
    """

    # Collect parameters
    name = config['name']
    
    if 'master' in config:
        master = config['master']
    elif 'ENRICH_SPARK_MASTER' in os.environ:
        master = os.environ['ENRICH_SPARK_MASTER']
    else:
        master = 'local'

    jar_packages = config.get('jar_packages',[])
    spark_config = config.get('spark_config',{})
    
    # start Spark application and get Spark session, logger and config
    spark, log, extra_config = start_spark(
        app_name=name, 
        master=master,
        jar_packages=jar_packages,
        spark_config=spark_config) 

    # Save it...
    sc = spark.sparkContext
    sqlsc = SQLContext(sc)

    config['spark'] = spark
    config['sc'] = sc 
    config['sqlsc'] = sqlsc    
    config['applicationid'] = sc.applicationId
    config['logger'] = log
    
    # log that main ETL job is starting
    log.warn('Pipeline is up-and-running')

    return

def stop_spark_session(config):
    
    spark = config['spark']
    log = config['log']
    
    log.warn('test_etl_job is finished')

    spark.stop()

    return None



"""
Dummy task 
"""

import time
import logging 
from enrichsdk.tasks import Task

logger = logging.getLogger("app") 

__all__ = ['DummyTask'] 


class DummyTask(Task): 
    """
    Backsup code and data as required 
    """
    NAME = "DummyTask" 

    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs) 
        self.name = "Dummy"
        self.testdata = {
            'data': {},
            'conf':{"enable_extra_args":False, 
                'args': {'sleep': 10}
        }}
        

        self.supported_extra_args = [
               {  
                  "name": "jobid",
                  "description": "Export JobID",
                  "type": "str",
                  "required": True,
                  "default": "22832",
               }
        ]

    def validate_args(self, what, state): 
        """
        Validate args. 
        
        Example::

            "args": {
                 "sleep": 10
            }
    
        """
        args = self.args
        fail = False
        msg = ""
        if "sleep" not in args: 
            fail = True 
            msg += "Argument passed does not have/is invalid sleep value\n" 

        if fail: 
            logger.error("Invalid params", 
                         extra=self.config.get_extra({
                             'transform': self.name,
                             'data': msg 
                             
                     }))
            raise Exception("Invalid params") 


    def run(self, state): 
        """
        Run the backup task 
        """

        # Get the number of seconds to sleep for
        sleep_val = self.args['sleep'] 

        # Check if the sleep argument is an integer
        assert isinstance(sleep_val, int)

        time.sleep(sleep_val)

        logger.debug("Sleep task complete.", 
                     extra=self.config.get_extra({
                         'transform': self.name, 
                     }))


import os
import sys
import json
import copy
import logging
import traceback
import humanize
from datetime import datetime
from dateutil import parser as date_parser
import logging

logger = logging.getLogger('app')

###############################################################
# Run History
###############################################################
class Run(object):
    """
    A single instance of command run. This class will mediate
    access to the command logs and metadata.

    """
    def __init__(self, rundir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rundir = rundir
        self.runid = os.path.basename(self.rundir)
        epoch = datetime.fromtimestamp(0)
        self.start_time = epoch
        self.end_time = epoch

        try:
            filename = os.path.join(self.rundir, "metadata.json")
            self.metadata = json.load(open(filename))
            schema = self.metadata.get('schema', 'unknown')
            command = self.metadata.get('command', {})

            for attr in ['host', 'name', 'status',
                         'description', 'cmdline',
                         'message', 'user', 'runid',
                         'start_time', 'end_time']:

                if schema  == 'standalone:command':
                    if attr in command:
                        if attr in ['start_time', 'end_time']:
                            value = date_parser.parse(command[attr])
                        else:
                            value = command[attr]
                        setattr(self, attr, value)
                    elif attr == 'user':
                        value = command['stats']['user']
                        setattr(self, attr, value)
                    elif attr == 'user':
                        value = command['stats']['user']
                        setattr(self, attr, value)
                else:
                    if attr in ['start_time', 'end_time']:
                        value = date_parser.parse(self.metadata[attr])
                    elif attr == 'user': 
                        value = self.metadata.get('username', "unknown")
                    elif attr == 'name': 
                        value = self.metadata.get('command', "unknown")                        
                    else:
                        value = self.metadata.get(attr, "unknown")

                    # Rename one attribute
                    setattr(self, attr, value)

            self.outputs = self.metadata.get('outputs', [])
        except:
            traceback.print_exc()
            for attr in ['host', 'name', 'status',
                         'description', 'cmdline'
                         'message', 'user']:
                setattr(self, attr, 'unknown')
            self.outputs = []

        logfile = self.metadata.get('log', 'log.json')
        self.logfile = os.path.join(self.rundir, logfile)
        self.log = open(self.logfile).readlines()

        self.duration = (self.end_time - self.start_time).total_seconds()

    def __str__(self):
        return self.name

    @property
    def username(self):
        return self.user
    
    def get_command(self):
        return self.name.replace('enrichcmd run ', '').strip()

    def get_cmdline(self):
        cmdline = copy.copy(self.cmdline)
        if cmdline[0].endswith('enrichcmd'):
            cmdline[0] = 'enrichcmd'
        return " ".join(cmdline)

class RunManager(object):
    """
    Command Run Management to enable access to
    pass runs of the commands.

    """
    def __init__(self, conf, *args, **kwargs):
        """

        Args:
            conf (dict): Only one parameter supported now 'command_root'
        """
        super().__init__(*args, **kwargs)
        self.runs = []
        self.conf = conf

        if 'command_root' in conf:
            command_roots = [conf['command_root']]
        elif 'command_roots' in conf:
            command_roots = conf['command_roots']
        else:
            raise Exception("Configuration should specify command_root or command_roots parameters")

        self.command_roots = []
        for p in command_roots:
            self.command_roots.append(os.path.abspath(p))

        # Collect rundirs from all directories...
        rundirs = []
        for p in self.command_roots:
            if not os.path.exists(p):
                logger.debug("Invalid command root directory: {}".format(p))
                continue

            rundirs.extend([
                {
                    'name': d,
                    'path': os.path.join(p, d)
                } for d in os.listdir(p)
            ])

        rundirs = sorted(rundirs, key=lambda r: r['name'], reverse=True)

        count = conf.get('history', 100)
        rundirs = rundirs[:count]
        errors = []
        for r in rundirs:
            try:
                runid = r['name']
                rundir = r['path']
                self.runs.append(Run(rundir))
            except:
                traceback.print_exc()
                errors.append(runid)

        if len(errors) > 0:
            logger.debug("Found errors while loading command history",
                         extra = {
                             'transform': 'Commands',
                             'data': str(errors)
                         })

    def get_runs(self, username=None):
        """
        Get runs for a given username
        """

        runs = [r for r in self.runs if ((username is None) or
                                         (r.user == username))]
        return runs

    def get_run(self, runid):

        run = None
        for p in self.command_roots:
            rundir = os.path.join(p, runid)
            if os.path.exists(rundir):
                try:
                    run = Run(rundir)
                except:
                    pass

        return run


    def show_history(self, n=10):

        runs = sorted(self.runs,
                      key=lambda r: r.start_time,
                      reverse=True)

        template = "{:10s}{:30s}{:10s}{:15s}{:20s}"
        print(template.format("Who","Description","Status","When","Name"))
        for r in runs[:n]:
            print(template.format(r.user,
                                  r.description[:27],
                                  r.status[:10],
                                  humanize.naturaltime(r.start_time),
                                  r.name))



"""
Transforms
----------

Modules to transform streaming records

"""
import os
import sys
import json
import logging
from datetime import datetime

logger = logging.getLogger('app')

from .db import RedisBackend

##############################################
# Base classes
##############################################
class TransformBase(object):
    """
    Base for a realtime feature engineering transform
    """
    def __init__(self, *args, **kwargs):
        self.version = "v1.0"
        self.name = self.display = self.__class__.__name__
        self.pipeline = None
        self.args = {}

    def configure(self, pipeline, args):

        self.pipeline = pipeline
        self.logger = pipeline.logger
        self.args = args

        # Override the transform name..
        self.enable  = self.args.get('enable', True)
        self.display = self.args.get('display', self.display)
        self.version = self.args.get('version', self.version)

    def initialize(self):
        pass

    def process(self, state):
        pass

    def __str__(self):
        return "{}:{}".format(self.display, self.version)

    def get_pipeline_extra(self):
        if self.pipeline is not None:
            return self.pipeline.get_extra()
        else:
            return {}

##############################################
# DB Access Instances
##############################################
class RedisTransformBase(TransformBase):
    """
    Base class for Redis Read/Write Transforms
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self):
        connparams = self.args.get('connection', {})
        self.backend = RedisBackend(config=connparams)
        self.backend.connect()

class RedisReadTransform(RedisTransformBase):
    """
    Read from Redis
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, state):

        key = self.args.get('key',None)
        if key is None:
            return state

        streamrow = state.get('streamrow')

        try:
            key = key % streamrow
        except:
            logger.exception("Cant compute the key")
            return state

        name = self.args.get('name', 'dbrow')
        dbrow = self.backend.get(key)
        state.put(name, dbrow)

        return state

class RedisWriteTransform(RedisTransformBase):
    """
    Write to Redis
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, state):

        key = self.args.get('key',None)
        if key is None:
            return state

        streamrow = state.get('streamrow')

        name = self.args.get('name', 'dbrow')
        dbrow = state.get(name)

        try:
            key = key % streamrow
        except:
            logger.exception("Cant compute the key from row")
            return state

        try:
            self.backend.post(key, dbrow)
        except:
            logger.exception("Unable to post to backend")
            pass


        return state


##############################################
# Metadata
##############################################
class MetaStartTransform(TransformBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, state):

        super().process(state)

        streamrow = state.get('streamrow')

        # Insert extra information
        if not '_meta_' in streamrow:
            streamrow['_meta_'] = {}

        streamrow['_meta_'].update({
            'start_ts': datetime.now().isoformat()
        })

        state.put('streamrow', streamrow)

        return state


class MetaEndTransform(TransformBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, state):

        super().process(state)

        streamrow = state.get('streamrow')

        extra = self.get_pipeline_extra()
        extra.update({
            'end_ts': datetime.now().isoformat()
        })

        if not '_meta_' in streamrow:
            streamrow['_meta_'] = {}

        streamrow['_meta_'].update(extra)

        return state

class LoggerTransform(TransformBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self):
        name = self.args.get('name', __name__)
        self.logger = logging.getLogger(name)

    def process(self, state):

        super().process(state)

        streamrow = state.get('streamrow')
        dbrow = state.get('dbrow',{})

        # Clean up the record before posting
        streamrow.pop('message',"")
        streamrow['_meta_'].pop('message',"")
        extra = self.args.get('extra',{})
        extra['data'] = {
            'streamrow': streamrow,
            'dbrow': dbrow
        }

        self.logger.debug("Message", extra=extra)

        return state



def has_permissions(user, orgname, role=None): 
    """
    Check whether user has permissions 
    """

    if not user.is_authenticated:
        return False 

    if user.is_superuser: 
        return True 

    orgroles = list(user.org_role_users.all())

    # Filter out all other organizations
    orgroles = [ o for o in orgroles if o.org.name == orgname] 

    # For the specific organization 
    roles = [o.name for o in orgroles] 
    
    if (('admin' in roles) or 
        (role is not None and role in roles)): 
        return True 

    # Denied
    return False 
